from typing import Optional, Union, List
from datetime import datetime
from concurrent.futures import ThreadPoolExecutor

import pandas as pd

from .. import _credentials
from .. import endpoints
from .. import utils

__all__ = [
    "get_batch_types",
    "get_batch_type_details",
    "get_batches",
    "update_batch_values",
    "update_batch_feature_values",
]


def get_batch_types(
    page: Optional[int] = None,
    page_size: Optional[int] = None,
    get_all_pages: bool = True,
    multithread_pages: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """
    Get the configured batch types from the OIAnalytics API

    Parameters
    ----------
    page: int, optional
        Page number to retrieve. If None, the first page will be retrieved.
        The argument is ignored if 'get_all_pages' is True.
    page_size: int, optional
        The size of each page to retrieve. By default, 20 elements are retrieved.
        The argument is ignored if 'get_all_pages' is True.
    get_all_pages: bool, default True
        If True, paging is ignored and all elements are retrieved.
    multithread_pages: bool, default False
        Only used when getting all pages. If True, pages are retrieved in multiple threads simultaneously.
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    A DataFrame listing batch types
    """

    # Init
    if get_all_pages is True:
        page = 0
        page_size = 500

    def get_page(page_num: int):
        page_response = endpoints.batches.get_batch_types(
            page=page_num, page_size=page_size, api_credentials=api_credentials
        )
        return page_response

    def parse_page(page_response: dict):
        page_df = pd.DataFrame(page_response["content"])

        # Expected columns if content is empty
        if page_df.shape[0] == 0:
            page_df = pd.DataFrame(columns=["id", "name"])

        page_df.set_index("id", inplace=True)
        return page_df

    # Query endpoint
    df = utils.concat_pages_to_dataframe(
        getter=get_page,
        parser=parse_page,
        page=page,
        get_all_pages=get_all_pages,
        multithread_pages=multithread_pages,
    )

    # Output
    return df


def get_batch_type_details(
    batch_type_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """
    Get details about a single batch type from the OIAnalytics API

    Parameters
    ----------
    batch_type_id: str
        The id of the batch type to be retrieved
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    A tuple of 3 DataFrames listing various properties of the batch type:
        - Steps
        - Data
        - Features
    """

    # Query endpoint
    response = endpoints.batches.get_batch_type_details(
        batch_type_id=batch_type_id, api_credentials=api_credentials
    )

    # Split content
    steps = response["steps"]
    data = response["dataList"]
    features = response["features"]

    # Format dataframes
    if len(steps) > 0:
        df_steps = pd.DataFrame(steps).set_index("id")
    else:
        df_steps = pd.DataFrame(columns=["id", "name", "localisationKeys"]).set_index(
            "id"
        )

    if len(data) > 0:
        df_data = pd.DataFrame(data).set_index("id")
    else:
        df_data = pd.DataFrame(
            columns=["id", "dataType", "reference", "description"]
        ).set_index("id")

    if len(features) > 0:
        df_features = pd.DataFrame(features).set_index("id")
    else:
        df_features = pd.DataFrame(columns=["id", "key"]).set_index("id")

    # Output
    return df_steps, df_data, df_features


def get_batches(
    batch_type_id: str,
    start_date: Union[str, datetime],
    end_date: Union[str, datetime],
    name: Optional[str] = None,
    features_value_ids: Optional[Union[str, List[str]]] = None,
    page: Optional[int] = None,
    page_size: Optional[int] = None,
    get_all_pages: bool = True,
    multithread_pages: bool = True,
    split_steps_and_values: bool = True,
    extract_from_localisation: Optional[str] = "value",
    expand_localisation: bool = True,
    extract_from_values: Optional[str] = "reference",
    expand_value_ids: bool = True,
    extract_from_features: Optional[str] = "value",
    expand_features: bool = True,
    append_unit_to_description: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """
    Get batch instances from the OIAnalytics API

    Parameters
    ----------
    batch_type_id: str
        The id of the batch type to be retrieved
    start_date: datetime or str
        The beginning of the period to be retrieved
    end_date: datetime or str
        The end of the period to be retrieved
    name: str, optional
        A string that should be contained by all batch names returned
    features_value_ids: str or list of str, optional
        Possibly multiple feature value ids each returned batch should match.
        If for a given feature multiple feature value ids are provided then a batch will be returned if it
        contains one of them.
    page: int, optional
        Page number to retrieve. If None, the first page will be retrieved.
        The argument is ignored if 'get_all_pages' is True.
    page_size: int, optional
        The size of each page to retrieve. By default, 20 elements are retrieved.
        The argument is ignored if 'get_all_pages' is True.
    get_all_pages: bool, default True
        If True, paging is ignored and all elements are retrieved.
    multithread_pages: bool, default False
        Only used when getting all pages. If True, pages are retrieved in multiple threads simultaneously.
    split_steps_and_values: bool, default True
        If True, the response will be split into two separate DataFrames instead of a single dense one.
    extract_from_localisation: {'id', 'value', None}, default 'value'
        What field should be extracted from localisation information. If None, the full dictionary is kept.
    expand_localisation: bool, default True
        Whether or not localisation information should be expanded into multiple columns.
    extract_from_values: {'id', 'reference', 'description', None}, default 'reference'
        What field should be extracted for naming data. If None, the full dictionary is kept.
    expand_value_ids: bool, default True
        Whether or not data should be expanded into multiple columns.
    extract_from_features: {'id', 'value', None}, default 'value'
        What field should be extracted for naming features. If None, the full dictionary is kept.
    expand_features: bool, default True
        Whether or not features should be expanded into multiple columns.
    append_unit_to_description: bool, default True
        Only used when 'extract_from_values' is 'description'. If True, the unit is added after the description.
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    If 'split_steps_and_values' is False, a single DataFrame containing batches and their information in dictionaries
    is returned.
    If 'split_steps_and_values' is True (default behaviour), a tuple of 2 DataFrames is returned:
        - Batch steps
        - Batch data and features
    """

    # Args validation
    if extract_from_localisation not in ["id", "value", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_localisation': {extract_from_localisation}"
        )

    if extract_from_localisation is None and expand_localisation is True:
        raise ValueError(
            "Localisation cannot be expanded if 'extract_from_values' is None"
        )

    if extract_from_values not in ["id", "reference", "description", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_values': {extract_from_values}"
        )

    if extract_from_values is None and expand_value_ids is True:
        raise ValueError("Values cannot be expanded if 'extract_from_values' is None")

    if extract_from_features not in ["id", "value", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_features': {extract_from_features}"
        )

    if extract_from_features is None and expand_features is True:
        raise ValueError(
            "Features cannot be expanded if 'extract_from_features' is None"
        )

    # Init
    if get_all_pages is True:
        page = 0
        page_size = 500

    def get_page(page_num: int):
        page_response = endpoints.batches.get_batches(
            batch_type_id=batch_type_id,
            start_date=start_date,
            end_date=end_date,
            name=name,
            features_value_ids=features_value_ids,
            page=page_num,
            page_size=page_size,
            api_credentials=api_credentials,
        )
        return page_response

    def parse_page(page_response: dict):
        page_df = pd.DataFrame(page_response["content"])

        # Expected columns if content is empty
        if page_df.shape[0] == 0:
            page_df = pd.DataFrame(
                columns=["id", "name", "timestamp", "steps", "values", "features"]
            )

        # Rename columns
        page_df.rename(
            columns={
                "id": "batch_id",
                "name": "batch_name",
                "timestamp": "batch_timestamp",
            },
            inplace=True,
        )

        # Parse dates
        page_df["batch_timestamp"] = pd.to_datetime(page_df["batch_timestamp"])

        # Set index
        page_df.set_index(["batch_id", "batch_name", "batch_timestamp"], inplace=True)
        return page_df

    # Query endpoint
    df = utils.concat_pages_to_dataframe(
        getter=get_page,
        parser=parse_page,
        page=page,
        get_all_pages=get_all_pages,
        multithread_pages=multithread_pages,
    )

    # Split steps and values
    if split_steps_and_values is True:
        # Split dataframe
        df_steps = df.drop(columns=["values", "features"])
        df_values = df.drop(columns="steps")

        # Format steps
        df_steps = df_steps.explode("steps")
        df_steps = utils.expand_dataframe_column(
            df_steps,
            "steps",
            add_prefix=False,
            expected_keys=["step", "start", "end", "localisation"],
        )
        df_steps = utils.expand_dataframe_column(
            df_steps, "step", expected_keys=["id", "name"]
        )

        df_steps["start"] = pd.to_datetime((df_steps["start"]))
        df_steps["end"] = pd.to_datetime((df_steps["end"]))

        if extract_from_localisation == "id":
            df_steps["localisation"] = df_steps["localisation"].apply(
                lambda full_loc: {loc["tagKey"]["id"]: loc["id"] for loc in full_loc}
            )
        elif extract_from_localisation == "value":
            df_steps["localisation"] = df_steps["localisation"].apply(
                lambda full_loc: {
                    loc["tagKey"]["key"]: loc["value"] for loc in full_loc
                }
            )

        if expand_localisation is True and extract_from_localisation is not None:
            df_steps = utils.expand_dataframe_column(
                df_steps, "localisation", add_prefix=False
            )

        df_steps.set_index(["step_id"], append=True, inplace=True)

        # Format values
        if extract_from_values == "id":
            df_values["values"] = df_values["values"].apply(
                lambda values: {val["data"]["id"]: val["value"] for val in values}
            )
        elif extract_from_values == "reference":
            df_values["values"] = df_values["values"].apply(
                lambda values: {
                    val["data"]["reference"]: val["value"] for val in values
                }
            )
        elif extract_from_values == "description":
            if append_unit_to_description is True:
                df_values["values"] = df_values["values"].apply(
                    lambda values: {
                        f'{val["data"]["description"]} ({val["unit"]["label"]})': val[
                            "value"
                        ]
                        for val in values
                    }
                )
            else:
                df_values["values"] = df_values["values"].apply(
                    lambda values: {
                        val["data"]["description"]: val["value"] for val in values
                    }
                )

        if expand_value_ids is True and extract_from_values is not None:
            df_values = utils.expand_dataframe_column(
                df_values, "values", add_prefix=False
            )

        if extract_from_features == "id":
            df_values["features"] = df_values["features"].apply(
                lambda features: {feat["tagKey"]["id"]: feat["id"] for feat in features}
            )
        elif extract_from_features == "value":
            df_values["features"] = df_values["features"].apply(
                lambda features: {
                    feat["tagKey"]["key"]: feat["value"] for feat in features
                }
            )

        if expand_features is True and extract_from_features is not None:
            df_values = utils.expand_dataframe_column(
                df_values, "features", add_prefix=False
            )

        # Output
        return df_steps, df_values
    else:
        return df


def update_batch_values(
    batch_type_id: str,
    data: Union[pd.Series, pd.DataFrame],
    unit_ids: Optional[dict] = None,
    batch_id_index_name: str = "batch_id",
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Init
    if unit_ids is None:
        unit_ids = {}

    if isinstance(data, pd.Series):
        data = pd.DataFrame(data)
    else:
        data = data.copy()

    data.index = data.index.get_level_values(batch_id_index_name)

    # Send each column
    def send_row(row_tuple: tuple):
        row = row_tuple[1]
        for col in row.index:
            endpoints.batches.update_batch_value(
                batch_type_id=batch_type_id,
                batch_id=str(row.name),
                data_id=col,
                value=row[col],
                unit_id=unit_ids.get(col, None),
                api_credentials=api_credentials,
            )

    with ThreadPoolExecutor() as pool:
        pool.map(send_row, data.iterrows())


def update_batch_feature_values(
    batch_type_id: str,
    data: Union[pd.Series, pd.DataFrame],
    batch_id_index_name: str = "batch_id",
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Init
    if isinstance(data, pd.Series):
        data = pd.DataFrame(data)
    else:
        data = data.copy()

    data.index = data.index.get_level_values(batch_id_index_name)

    # Send each column
    def send_row(row_tuple: tuple):
        row = row_tuple[1]
        for col in row.index:
            endpoints.batches.update_batch_feature_value(
                batch_type_id=batch_type_id,
                batch_id=str(row.name),
                feature_id=col,
                value=row[col],
                api_credentials=api_credentials,
            )

    with ThreadPoolExecutor() as pool:
        pool.map(send_row, data.iterrows())



from typing import Optional, Union, List
from datetime import datetime

import pandas as pd

from .. import _credentials
from .. import endpoints
from .. import utils

__all__ = ["get_event_types", "get_events"]


def get_event_types(
    page: Optional[int] = None,
    page_size: Optional[int] = None,
    get_all_pages: bool = True,
    multithread_pages: bool = True,
    extract_from_tag_keys: Optional[str] = "value",
    extract_from_context: Optional[str] = "value",
    expand_context: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """
    Get the configured event types from the OIAnalytics API

    Parameters
    ----------
    page: int, optional
        Page number to retrieve. If None, the first page will be retrieved.
        The argument is ignored if 'get_all_pages' is True.
    page_size: int, optional
        The size of each page to retrieve. By default, 20 elements are retrieved.
        The argument is ignored if 'get_all_pages' is True.
    get_all_pages: bool, default True
        If True, paging is ignored and all elements are retrieved.
    multithread_pages: bool, default False
        Only used when getting all pages. If True, pages are retrieved in multiple threads simultaneously.
    extract_from_tag_keys: {'id', 'value', 'dict', None}, default 'value'
        What field should be extracted from tag keys information.
        If 'dict', a dictionary in the form of {id: value} is built. If None, the full dictionary is kept.
    extract_from_context: {'id', 'value', None}, default 'value'
        What field should be extracted from context information. If None, the full dictionary is kept.
    expand_context: bool, default True
        Whether or not the context should be expanded into multiple columns.
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    A DataFrame listing event types
    """

    # Args validation
    if extract_from_tag_keys not in ["id", "value", "dict", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_tag_keys': {extract_from_tag_keys}"
        )

    if extract_from_context not in ["id", "value", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_context': {extract_from_context}"
        )

    if expand_context is True and extract_from_context is None:
        raise ValueError("Context cannot be expanded if 'extract_from_context' is None")

    # Init
    if get_all_pages is True:
        page = 0
        page_size = 500

    def get_page(page_num: int):
        page_response = endpoints.events.get_event_types(
            page=page_num, page_size=page_size, api_credentials=api_credentials
        )
        return page_response

    def parse_page(page_response: dict):
        page_df = pd.DataFrame(page_response["content"])

        # Expected columns if content is empty
        if page_df.shape[0] == 0:
            page_df = pd.DataFrame(columns=["id", "name", "tagKeys", "tagContext"])

        # Extract from tag keys
        if extract_from_tag_keys == "id":
            page_df["tagKeys"] = page_df["tagKeys"].apply(
                lambda l: [tk["id"] for tk in l]
            )
        elif extract_from_tag_keys == "value":
            page_df["tagKeys"] = page_df["tagKeys"].apply(
                lambda l: [tk["key"] for tk in l]
            )
        elif extract_from_tag_keys == "dict":
            page_df["tagKeys"] = page_df["tagKeys"].apply(
                lambda l: {tk["id"]: tk["key"] for tk in l}
            )

        # Extract from context
        if extract_from_context == "id":
            page_df["tagContext"] = page_df["tagContext"].apply(
                lambda context: {
                    c["tagKey"]["id"]: [atv["id"] for atv in c["accessibleTagValues"]]
                    for c in context
                }
            )
        elif extract_from_context == "value":
            page_df["tagContext"] = page_df["tagContext"].apply(
                lambda context: {
                    c["tagKey"]["key"]: [
                        atv["value"] for atv in c["accessibleTagValues"]
                    ]
                    for c in context
                }
            )

        if expand_context is True and extract_from_context is not None:
            page_df = utils.expand_dataframe_column(
                page_df, "tagContext", add_prefix=False
            )

        page_df.set_index("id", inplace=True)
        return page_df

    # Query endpoint
    df = utils.concat_pages_to_dataframe(
        getter=get_page,
        parser=parse_page,
        page=page,
        get_all_pages=get_all_pages,
        multithread_pages=multithread_pages,
    )

    # Output
    return df


def get_events(
    event_type_id: str,
    start_date: Union[str, datetime],
    end_date: Union[str, datetime],
    description: Optional[str] = None,
    features_value_ids: Optional[Union[str, List[str]]] = None,
    page: Optional[int] = None,
    page_size: Optional[int] = None,
    get_all_pages: bool = True,
    multithread_pages: bool = True,
    expand_event_type: bool = True,
    extract_from_features: Optional[str] = "value",
    expand_features: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """
    Get event instances from the OIAnalytics API

    Parameters
    ----------
    event_type_id: str
        The id of the event type to be retrieved
    start_date: datetime or str
        The beginning of the period to be retrieved
    end_date: datetime or str
        The end of the period to be retrieved
    description: str, optional
        A string that should be contained by all events description returned
    features_value_ids: str or list of str, optional
        Possibly multiple feature value ids each returned event should match.
        If for a given feature multiple feature value ids are provided than a event will be returned if it
        contains one of them.
    page: int, optional
        Page number to retrieve. If None, the first page will be retrieved.
        The argument is ignored if 'get_all_pages' is True.
    page_size: int, optional
        The size of each page to retrieve. By default, 20 elements are retrieved.
        The argument is ignored if 'get_all_pages' is True.
    get_all_pages: bool, default True
        If True, paging is ignored and all elements are retrieved.
    multithread_pages: bool, default False
        Only used when getting all pages. If True, pages are retrieved in multiple threads simultaneously.
    expand_event_type: bool, default True
        Whether or not the event type information should be expanded into multiple columns.
    extract_from_features: {'id', 'value', None}, default 'value'
        What field should be extracted from features information. If None, the full dictionary is kept.
    expand_features: bool, default True
        Whether or not the features should be expanded into multiple columns.
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    A DataFrame containing event dates, data and features values
    """

    # Args validation
    if extract_from_features not in ["id", "value", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_features': {extract_from_features}"
        )

    if expand_features is True and extract_from_features is None:
        raise ValueError(
            "Features cannot be expanded if 'extract_from_features' is None"
        )

    # Init
    if get_all_pages is True:
        page = 0
        page_size = 500

    def get_page(page_num: int):
        page_response = endpoints.events.get_events(
            event_type_id=event_type_id,
            start_date=start_date,
            end_date=end_date,
            description=description,
            features_value_ids=features_value_ids,
            page=page_num,
            page_size=page_size,
            api_credentials=api_credentials,
        )
        return page_response

    def parse_page(page_response: dict):
        page_df = pd.DataFrame(page_response["content"])

        if page_df.shape[0] == 0:
            page_df = pd.DataFrame(
                columns=[
                    "eventType",
                    "id",
                    "description",
                    "start",
                    "end",
                    "duration",
                    "values",
                    "features",
                ]
            )

        # Format dataframe
        if expand_event_type is True:
            page_df = utils.expand_dataframe_column(
                page_df, "eventType", expected_keys=["id", "name"]
            )

        if extract_from_features == "id":
            page_df["features"] = page_df["features"].apply(
                lambda fl: {f["tagKey"]["id"]: f["id"] for f in fl}
            )
        elif extract_from_features == "value":
            page_df["features"] = page_df["features"].apply(
                lambda fl: {f["tagKey"]["key"]: f["value"] for f in fl}
            )

        if expand_features is True and extract_from_features is not None:
            page_df = utils.expand_dataframe_column(
                page_df, "features", add_prefix=False
            )

        page_df.set_index("id", inplace=True)

        # Output
        return page_df

    # Query endpoint
    df = utils.concat_pages_to_dataframe(
        getter=get_page,
        parser=parse_page,
        page=page,
        get_all_pages=get_all_pages,
        multithread_pages=multithread_pages,
    )

    # Output
    return df



from typing import Optional, Union, List
from datetime import datetime, timedelta, timezone
import dateutil

from .. import _dtos
from ... import api

__all__ = ["TestModelExecution"]


class TestModelExecution(_dtos.ModelExecution):
    def __init__(
        self,
        test_mode: bool = True,
        last_successful_execution_instant: datetime = datetime.now(timezone.utc)
        - timedelta(hours=1),
        current_execution_instant: datetime = datetime.now(timezone.utc),
        data_exchange_mode: str = "FREE_FROM_API",
        single_observation_type: str = "time",
        single_observation_overlapping_period: timedelta = timedelta(hours=1),
        single_observation_aggregation_period: str = "PT10M",
        single_observation_batch_type_id: str = "single_obs_batch_type_id",
        single_observation_batch_type_reference: str = "single_obs_batch_type_reference",
    ):
        # Data exchange mode
        if data_exchange_mode == "FREE_FROM_API":
            single_obs_context = None
        elif data_exchange_mode == "SINGLE_OBSERVATION":
            if single_observation_type == "time":
                single_obs_context = _dtos.SingleObservationTimeContext(
                    type="time",
                    overlappingPeriod=single_observation_overlapping_period,
                    aggregationPeriod=single_observation_aggregation_period,
                )
            elif single_observation_type == "batch":
                single_obs_context = _dtos.SingleObservationBatchContext(
                    type="batch",
                    overlappingPeriod=single_observation_overlapping_period,
                    batchPredicate=_dtos.BatchPredicateValue(
                        type="batch-predicate",
                        batchType=_dtos.BatchPredicateBatchType(
                            type="batch-type",
                            id=single_observation_batch_type_id,
                            reference=single_observation_batch_type_reference,
                        ),
                        featureFilters=[],
                    ),
                )
            else:
                raise ValueError(
                    f"{single_observation_type} is not valid for 'single_observation_type'"
                )
        else:
            raise ValueError(
                f"{data_exchange_mode} is not valid for 'data_exchange_mode'"
            )

        # Dates parsing and formatting
        if isinstance(last_successful_execution_instant, str):
            last_execution = dateutil.parser.parse(last_successful_execution_instant)
        else:
            last_execution = last_successful_execution_instant
        if isinstance(current_execution_instant, str):
            current_execution = dateutil.parser.parse(current_execution_instant)
        else:
            current_execution = current_execution_instant

        # Init
        super().__init__(
            testMode=test_mode,
            credentials=None,
            lastSuccessfulExecutionInstant=last_execution.astimezone(timezone.utc),
            currentExecutionInstant=current_execution.astimezone(timezone.utc),
            pythonModelInstance=_dtos.ModelInstance(
                id="test_instance_id",
                trigger=_dtos.CronTrigger(type="cron-trigger", cron="0 0 * * * *"),
                active=False,
                dataExchangeMode=data_exchange_mode,
                singleObservationContext=single_obs_context,
                inputParameters=[],
                outputParameters=[],
            ),
        )

    @classmethod
    def from_model_execution(cls, model_execution: _dtos.ModelExecution):
        test_model_execution = cls()
        test_model_execution.testMode = model_execution.testMode
        test_model_execution.credentials = model_execution.credentials
        test_model_execution.lastSuccessfulExecutionInstant = (
            model_execution.lastSuccessfulExecutionInstant
        )
        test_model_execution.currentExecutionInstant = (
            model_execution.currentExecutionInstant
        )
        test_model_execution.pythonModelInstance = model_execution.pythonModelInstance
        return test_model_execution

    def add_single_observation_feature_filter(
        self,
        feature_id: str = "single_obs_feature_id",
        feature_reference: str = "single_obs_feature_reference",
    ):
        self.pythonModelInstance.singleObservationContext.batchPredicate.featureFilters.append(
            _dtos.BatchTagValue(
                type="batch-tag-value", id=feature_id, reference=feature_reference
            )
        )

    def add_string_input(
        self,
        source_code_name: str = "string_input",
        value: Optional[str] = "String input",
    ):
        self.pythonModelInstance.inputParameters.append(
            _dtos.StringParameter(
                type="STRING", sourceCodeName=source_code_name, value=value
            )
        )

    def add_boolean_input(
        self, source_code_name: str = "boolean_input", value: Optional[bool] = True
    ):
        self.pythonModelInstance.inputParameters.append(
            _dtos.BooleanParameter(
                type="BOOLEAN", sourceCodeName=source_code_name, value=value
            )
        )

    def add_numeric_input(
        self,
        source_code_name: str = "numeric_input",
        value: Optional[Union[int, float]] = 42,
    ):
        self.pythonModelInstance.inputParameters.append(
            _dtos.NumericParameter(
                type="NUMERIC", sourceCodeName=source_code_name, value=value
            )
        )

    def add_file_input(
        self,
        source_code_name: str = "file_input",
        value: Optional[str] = "file_id",
    ):
        self.pythonModelInstance.inputParameters.append(
            _dtos.FileParameter(
                type="FILE", sourceCodeName=source_code_name, value=value
            )
        )

    def add_period_input(
        self,
        source_code_name: str = "period_input",
        value: Optional[str] = "PT12H",
    ):
        self.pythonModelInstance.inputParameters.append(
            _dtos.PeriodParameter(
                type="PERIOD", sourceCodeName=source_code_name, value=value
            )
        )

    def add_duration_input(
        self,
        source_code_name: str = "duration_input",
        value: Optional[int] = 86_400_000,
    ):
        self.pythonModelInstance.inputParameters.append(
            _dtos.DurationParameter(
                type="DURATION", sourceCodeName=source_code_name, value=value
            )
        )

    def add_instant_input(
        self,
        source_code_name: str = "instant_input",
        value: Optional[str] = api.utils.get_zulu_isoformat(datetime.now()),
    ):
        self.pythonModelInstance.inputParameters.append(
            _dtos.InstantParameter(
                type="INSTANT", sourceCodeName=source_code_name, value=value
            )
        )

    def add_unit_input(
        self,
        source_code_name: str = "unit_input",
        unit_id: Optional[str] = "unit_id",
        unit_label: str = "Unit label",
    ):
        if unit_id is None:
            unit_value = None
        else:
            unit_value = _dtos.UnitValue(type="unit", id=unit_id, reference=unit_label)

        self.pythonModelInstance.inputParameters.append(
            _dtos.UnitParameter(
                type="UNIT", sourceCodeName=source_code_name, value=unit_value
            )
        )

    def add_measurement_input(
        self,
        source_code_name: str = "measurement_input",
        measurement_id: Optional[str] = "measurement_id",
        measurement_reference: str = "Measurement reference",
    ):
        if measurement_id is None:
            measurement_value = None
        else:
            measurement_value = _dtos.MeasurementValue(
                type="measurement", id=measurement_id, reference=measurement_reference
            )

        self.pythonModelInstance.inputParameters.append(
            _dtos.MeasurementParameter(
                type="MEASUREMENT",
                sourceCodeName=source_code_name,
                value=measurement_value,
            )
        )

    def add_tagkey_input(
        self,
        source_code_name: str = "tagkey_input",
        tagkey_id: Optional[str] = "tagkey_id",
        tagkey_reference: str = "Tag key reference",
    ):
        if tagkey_id is None:
            tagkey_value = None
        else:
            tagkey_value = _dtos.TagKeyValue(
                type="tag-key", id=tagkey_id, reference=tagkey_reference
            )

        self.pythonModelInstance.inputParameters.append(
            _dtos.TagKeyParameter(
                type="TAG_KEY",
                sourceCodeName=source_code_name,
                value=tagkey_value,
            )
        )

    def add_tagvalue_input(
        self,
        source_code_name: str = "tagvalue_input",
        tagvalue_id: Optional[str] = "tagvalue_id",
        tagvalue_reference: str = "Tag value reference",
    ):
        if tagvalue_id is None:
            tagvalue_value = None
        else:
            tagvalue_value = _dtos.TagValueValue(
                type="tag-key", id=tagvalue_id, reference=tagvalue_reference
            )

        self.pythonModelInstance.inputParameters.append(
            _dtos.TagValueParameter(
                type="TAG_VALUE",
                sourceCodeName=source_code_name,
                value=tagvalue_value,
            )
        )

    def add_data_input(
        self,
        source_code_name: str = "data_input",
        data_id: Optional[str] = "data_id",
        data_reference: str = "data_reference",
        data_type: str = "STORED_CONTINUOUS",
        unit_id: str = "unit_id",
        unit_label: str = "Unit label",
        aggregation_function: Optional[str] = None,
    ):
        if data_id is None:
            data_value = None
        else:
            data_value = _dtos.DataValue(
                type="data",
                id=data_id,
                reference=data_reference,
                dataType=data_type,
                unit=_dtos.DataUnit(id=unit_id, label=unit_label),
                aggregationFunction=aggregation_function,
            )

        self.pythonModelInstance.inputParameters.append(
            _dtos.DataParameter(
                type="DATA",
                sourceCodeName=source_code_name,
                value=data_value,
            )
        )

    def add_timedata_input(
        self,
        source_code_name: str = "timedata_input",
        data_id: Optional[str] = "timedata_id",
        data_reference: str = "timedata_reference",
        unit_id: str = "unit_id",
        unit_label: str = "Unit label",
        aggregation_function: Optional[str] = None,
    ):
        if data_id is None:
            data_value = None
        else:
            data_value = _dtos.DataValue(
                type="data",
                id=data_id,
                reference=data_reference,
                dataType="STORED_CONTINUOUS",
                unit=_dtos.DataUnit(id=unit_id, label=unit_label),
                aggregationFunction=aggregation_function,
            )

        self.pythonModelInstance.inputParameters.append(
            _dtos.StoredContinuousDataParameter(
                type="STORED_CONTINUOUS_DATA",
                sourceCodeName=source_code_name,
                value=data_value,
            )
        )

    def add_batchdata_input(
        self,
        source_code_name: str = "batchdata_input",
        data_id: Optional[str] = "batchdata_id",
        data_reference: str = "batchdata_reference",
        unit_id: str = "unit_id",
        unit_label: str = "Unit label",
        aggregation_function: Optional[str] = None,
        batch_type_id: str = "batch_type_id",
        batch_type_name: str = "Batch type name",
    ):
        if data_id is None:
            data_value = None
        else:
            data_value = _dtos.BatchDataValue(
                type="batch-data",
                id=data_id,
                reference=data_reference,
                dataType="STORED_BATCH",
                unit=_dtos.DataUnit(id=unit_id, label=unit_label),
                aggregationFunction=aggregation_function,
                batchType=_dtos.BatchType(id=batch_type_id, name=batch_type_name),
            )

        self.pythonModelInstance.inputParameters.append(
            _dtos.StoredBatchDataParameter(
                type="STORED_BATCH_DATA",
                sourceCodeName=source_code_name,
                value=data_value,
            )
        )

    def add_computed_timedata_input(
        self,
        source_code_name: str = "computed_timedata_input",
        data_id: Optional[str] = "computed_timedata_id",
        data_reference: str = "computed_timedata_reference",
        unit_id: str = "unit_id",
        unit_label: str = "Unit label",
        aggregation_function: Optional[str] = None,
    ):
        if data_id is None:
            data_value = None
        else:
            data_value = _dtos.DataValue(
                type="data",
                id=data_id,
                reference=data_reference,
                dataType="COMPUTED_CONTINUOUS",
                unit=_dtos.DataUnit(id=unit_id, label=unit_label),
                aggregationFunction=aggregation_function,
            )

        self.pythonModelInstance.inputParameters.append(
            _dtos.ComputedContinuousDataParameter(
                type="COMPUTED_CONTINUOUS_DATA",
                sourceCodeName=source_code_name,
                value=data_value,
            )
        )

    def add_computed_batchdata_input(
        self,
        source_code_name: str = "computed_batchdata_input",
        data_id: Optional[str] = "computed_batchdata_id",
        data_reference: str = "computed_batchdata_reference",
        unit_id: str = "unit_id",
        unit_label: str = "Unit label",
        aggregation_function: Optional[str] = None,
        batch_type_id: str = "batch_type_id",
        batch_type_name: str = "Batch type name",
    ):
        if data_id is None:
            data_value = None
        else:
            data_value = _dtos.BatchDataValue(
                type="batch-data",
                id=data_id,
                reference=data_reference,
                dataType="COMPUTED_BATCH",
                unit=_dtos.DataUnit(id=unit_id, label=unit_label),
                aggregationFunction=aggregation_function,
                batchType=_dtos.BatchType(id=batch_type_id, name=batch_type_name),
            )

        self.pythonModelInstance.inputParameters.append(
            _dtos.ComputedBatchDataParameter(
                type="COMPUTED_BATCH_DATA",
                sourceCodeName=source_code_name,
                value=data_value,
            )
        )

    def add_batch_timedata_input(
        self,
        source_code_name: str = "batch_timedata_input",
        data_id: Optional[str] = "batch_timedata_id",
        data_reference: str = "batch_timedata_reference",
        unit_id: str = "unit_id",
        unit_label: str = "Unit label",
        aggregation_function: Optional[str] = None,
        batch_type_id: str = "batch_type_id",
        batch_type_name: str = "Batch type name",
    ):
        if data_id is None:
            data_value = None
        else:
            data_value = _dtos.BatchDataValue(
                type="batch-data",
                id=data_id,
                reference=data_reference,
                dataType="BATCH_TIME",
                unit=_dtos.DataUnit(id=unit_id, label=unit_label),
                aggregationFunction=aggregation_function,
                batchType=_dtos.BatchType(id=batch_type_id, name=batch_type_name),
            )

        self.pythonModelInstance.inputParameters.append(
            _dtos.BatchTimeDataParameter(
                type="BATCH_TIME_DATA",
                sourceCodeName=source_code_name,
                value=data_value,
            )
        )

    def add_batch_predicate_input(
        self,
        source_code_name: str = "batch_predicate_input",
        batch_type_id: Optional[str] = "batch_type_id",
        batch_type_reference: str = "Batch type reference",
        feature_value_ids: Optional[List[str]] = None,
        feature_value_references: Optional[List[str]] = None,
    ):
        if batch_type_id is None:
            batch_predicate_value = None
        else:
            # Build feature filters
            if feature_value_ids is None and feature_value_references is None:
                feature_filters = []
            elif feature_value_ids is not None and feature_value_references is not None:
                if len(feature_value_ids) != len(feature_value_references):
                    raise ValueError(
                        "'feature_value_ids' and 'feature_value_references' should have the same length"
                    )
                else:
                    feature_filters = [
                        _dtos.BatchTagValue(
                            type="batch-tag-value",
                            id=feature_value_ids[i],
                            reference=feature_value_references[i],
                        )
                        for i in range(len(feature_value_ids))
                    ]
            elif feature_value_ids is not None:
                feature_filters = [
                    _dtos.BatchTagValue(
                        type="batch-tag-value",
                        id=fid,
                        reference=f"{fid}_reference",
                    )
                    for fid in feature_value_ids
                ]
            else:
                feature_filters = [
                    _dtos.BatchTagValue(
                        type="batch-tag-value",
                        id=f"{fref}_id",
                        reference=fref,
                    )
                    for fref in feature_value_references
                ]

            # Build batch predicate
            batch_predicate_value = _dtos.BatchPredicateValue(
                type="batch-predicate",
                batchType=_dtos.BatchPredicateBatchType(
                    type="batch-type",
                    id=batch_type_id,
                    reference=batch_type_reference,
                ),
                featureFilters=feature_filters,
            )

        self.pythonModelInstance.inputParameters.append(
            _dtos.BatchPredicateParameter(
                type="BATCH_PREDICATE",
                sourceCodeName=source_code_name,
                value=batch_predicate_value,
            )
        )

    def add_batch_structure_input(
        self,
        source_code_name: str = "batch_structure_input",
        batch_type_id: Optional[str] = "batch_type_id",
        batch_type_reference: str = "Batch type reference",
    ):
        if batch_type_id is None:
            batch_structure_value = None
        else:
            batch_structure_value = _dtos.BatchStructureValue(
                type="batch-type", id=batch_type_id, reference=batch_type_reference
            )

        self.pythonModelInstance.inputParameters.append(
            _dtos.BatchStructureParameter(
                type="BATCH_STRUCTURE",
                sourceCodeName=source_code_name,
                value=batch_structure_value,
            )
        )

    def add_batch_tagkey_input(
        self,
        source_code_name: str = "batch_tagkey_input",
        tagkey_id: Optional[str] = "batch_tagkey_id",
        tagkey_reference: str = "Batch tag key reference",
    ):
        if tagkey_id is None:
            batch_tagkey_value = None
        else:
            batch_tagkey_value = _dtos.TagKeyValue(
                type="batch-tag-key", id=tagkey_id, reference=tagkey_reference
            )

        self.pythonModelInstance.inputParameters.append(
            _dtos.BatchTagKeyParameter(
                type="BATCH_TAG_KEY",
                sourceCodeName=source_code_name,
                value=batch_tagkey_value,
            )
        )

    def add_batch_tagvalue_input(
        self,
        source_code_name: str = "batch_tagvalue_input",
        tagvalue_id: Optional[str] = "batch_tagvalue_id",
        tagvalue_reference: str = "Batch tag value reference",
    ):
        if tagvalue_id is None:
            batch_tagvalue_value = None
        else:
            batch_tagvalue_value = _dtos.TagValueValue(
                type="batch-tag-value", id=tagvalue_id, reference=tagvalue_reference
            )

        self.pythonModelInstance.inputParameters.append(
            _dtos.BatchTagValueParameter(
                type="BATCH_TAG_VALUE",
                sourceCodeName=source_code_name,
                value=batch_tagvalue_value,
            )
        )

    def add_event_type_input(
        self,
        source_code_name: str = "event_type_input",
        event_type_id: Optional[str] = "event_type_id",
        event_type_reference: str = "Event type reference",
    ):
        if event_type_id is None:
            event_type_value = None
        else:
            event_type_value = _dtos.TagKeyValue(
                type="event-type", id=event_type_id, reference=event_type_reference
            )

        self.pythonModelInstance.inputParameters.append(
            _dtos.EventTypeParameter(
                type="EVENT_TYPE",
                sourceCodeName=source_code_name,
                value=event_type_value,
            )
        )

    def add_event_tagkey_input(
        self,
        source_code_name: str = "event_tagkey_input",
        tagkey_id: Optional[str] = "event_tagkey_id",
        tagkey_reference: str = "Event tag key reference",
    ):
        if tagkey_id is None:
            event_tagkey_value = None
        else:
            event_tagkey_value = _dtos.EventTagKeyValue(
                type="event-tag-key", id=tagkey_id, reference=tagkey_reference
            )

        self.pythonModelInstance.inputParameters.append(
            _dtos.EventTagKeyParameter(
                type="EVENT_TAG_KEY",
                sourceCodeName=source_code_name,
                value=event_tagkey_value,
            )
        )

    def add_event_tagvalue_input(
        self,
        source_code_name: str = "event_tagvalue_input",
        tagvalue_id: Optional[str] = "event_tagvalue_id",
        tagvalue_reference: str = "Event tag value reference",
    ):
        if tagvalue_id is None:
            event_tagvalue_value = None
        else:
            event_tagvalue_value = _dtos.EventTagValueValue(
                type="event-tag-value", id=tagvalue_id, reference=tagvalue_reference
            )

        self.pythonModelInstance.inputParameters.append(
            _dtos.EventTagValueParameter(
                type="EVENT_TAG_VALUE",
                sourceCodeName=source_code_name,
                value=event_tagvalue_value,
            )
        )

    def add_data_output(
        self,
        source_code_name: str = "data_output",
        data_id: Optional[str] = "data_id",
        data_reference: str = "data_reference",
        data_type: str = "STORED_CONTINUOUS",
        unit_id: str = "unit_id",
        unit_label: str = "Unit label",
        aggregation_function: Optional[str] = None,
    ):
        if data_id is None:
            data_value = None
        else:
            data_value = _dtos.DataValue(
                type="data",
                id=data_id,
                reference=data_reference,
                dataType=data_type,
                unit=_dtos.DataUnit(id=unit_id, label=unit_label),
                aggregationFunction=aggregation_function,
            )

        self.pythonModelInstance.outputParameters.append(
            _dtos.DataParameter(
                type="DATA",
                sourceCodeName=source_code_name,
                value=data_value,
            )
        )



from typing import Union, List, Optional
from datetime import datetime, timedelta

import pandas as pd
from pydantic import BaseModel, validator, StrictStr, StrictBool, StrictInt, StrictFloat

from .. import api

__all__ = [
    "ModelExecution",
    "get_default_model_execution",
    "set_default_model_execution",
]


# Credentials
class BasicAuthCredentials(BaseModel):
    baseUrl: StrictStr
    login: StrictStr
    pwd: StrictStr

    def to_object(self):
        return api.OIAnalyticsAPICredentials(
            base_url=self.baseUrl, login=self.login, pwd=self.pwd
        )


class TokenCredentials(BaseModel):
    baseUrl: StrictStr
    token: StrictStr

    def to_object(self):
        return api.OIAnalyticsAPICredentials(base_url=self.baseUrl, token=self.token)


Credentials = Union[BasicAuthCredentials, TokenCredentials]


# Triggers
class CronTrigger(BaseModel):
    type: StrictStr
    cron: StrictStr

    @validator("type")
    def check_type(cls, value):
        if value != "cron-trigger":
            raise ValueError("Trigger type should be 'cron-trigger'")
        return value


Trigger = Union[CronTrigger]


# Model inputs & outputs
class StringParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[StrictStr]

    @validator("type")
    def validate_type(cls, value):
        if value != "STRING":
            raise ValueError("'type' should be 'STRING'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value


class BooleanParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[StrictBool]

    @validator("type")
    def validate_type(cls, value):
        if value != "BOOLEAN":
            raise ValueError("'type' should be 'BOOLEAN'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value


class NumericParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[Union[StrictInt, StrictFloat]]

    @validator("type")
    def validate_type(cls, value):
        if value != "NUMERIC":
            raise ValueError("'type' should be 'NUMERIC'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value


class FileParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[StrictStr]

    @validator("type")
    def validate_type(cls, value):
        if value != "FILE":
            raise ValueError("'type' should be 'FILE'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value


class PeriodParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[str]

    @validator("type")
    def validate_type(cls, value):
        if value != "PERIOD":
            raise ValueError("'type' should be 'PERIOD'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value


class DurationParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[timedelta]

    @validator("type")
    def validate_type(cls, value):
        if value != "DURATION":
            raise ValueError("'type' should be 'DURATION'")
        return value

    @validator("value", pre=True)
    def parse_value(cls, value):
        if value is None:
            return None
        else:
            return pd.Timedelta(value, unit="ms").to_pytimedelta()

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value


class InstantParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[datetime]

    @validator("type")
    def validate_type(cls, value):
        if value != "INSTANT":
            raise ValueError("'type' should be 'INSTANT'")
        return value

    @validator("value", pre=True)
    def parse_value(cls, value):
        if value is None:
            return None
        else:
            return pd.to_datetime(value).to_pydatetime()

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value


class UnitValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr


class UnitParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[UnitValue]

    @validator("type")
    def validate_type(cls, value):
        if value != "UNIT":
            raise ValueError("'type' should be 'UNIT'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class MeasurementValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr


class MeasurementParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[MeasurementValue]

    @validator("type")
    def validate_type(cls, value):
        if value != "MEASUREMENT":
            raise ValueError("'type' should be 'MEASUREMENT'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class TagKeyValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr


class TagKeyParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[TagKeyValue]

    @validator("type")
    def validate_type(cls, value):
        if value != "TAG_KEY":
            raise ValueError("'type' should be 'TAG_KEY'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class TagValueValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr


class TagValueParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[TagValueValue]

    @validator("type")
    def validate_type(cls, value):
        if value != "TAG_VALUE":
            raise ValueError("'type' should be 'TAG_VALUE'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class DataUnit(BaseModel):
    id: StrictStr
    label: StrictStr


class DataValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr
    dataType: StrictStr
    unit: DataUnit
    aggregationFunction: Optional[StrictStr]


class DataParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[DataValue]

    @validator("type")
    def validate_type(cls, value):
        if value != "DATA":
            raise ValueError("'type' should be 'DATA'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class StoredContinuousDataParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[DataValue]

    @validator("type")
    def validate_type(cls, value):
        if value != "STORED_CONTINUOUS_DATA":
            raise ValueError("'type' should be 'STORED_CONTINUOUS_DATA'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class BatchType(BaseModel):
    id: StrictStr
    name: StrictStr


class BatchDataValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr
    dataType: StrictStr
    unit: DataUnit
    aggregationFunction: Optional[StrictStr]
    batchType: BatchType


class StoredBatchDataParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[BatchDataValue]

    @validator("type")
    def validate_type(cls, value):
        if value != "STORED_BATCH_DATA":
            raise ValueError("'type' should be 'STORED_BATCH_DATA'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class ComputedContinuousDataParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[DataValue]

    @validator("type")
    def validate_type(cls, value):
        if value != "COMPUTED_CONTINUOUS_DATA":
            raise ValueError("'type' should be 'COMPUTED_CONTINUOUS_DATA'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class ComputedBatchDataParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[BatchDataValue]

    @validator("type")
    def validate_type(cls, value):
        if value != "COMPUTED_BATCH_DATA":
            raise ValueError("'type' should be 'COMPUTED_BATCH_DATA'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class BatchTimeDataParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[BatchDataValue]

    @validator("type")
    def validate_type(cls, value):
        if value != "BATCH_TIME_DATA":
            raise ValueError("'type' should be 'BATCH_TIME_DATA'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class BatchPredicateBatchType(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr


class BatchTagValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr

    @validator("type")
    def check_type(cls, value):
        if value != "batch-tag-value":
            raise ValueError("Batch tag value type should be 'batch-tag-value'")
        return value


class BatchPredicateValue(BaseModel):
    type: StrictStr
    batchType: BatchPredicateBatchType
    featureFilters: List[BatchTagValue]

    @validator("type")
    def check_type(cls, value):
        if value != "batch-predicate":
            raise ValueError("Batch predicate type should be 'batch-predicate'")
        return value


class BatchPredicateParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[BatchPredicateValue]

    @validator("type")
    def validate_type(cls, value):
        if value != "BATCH_PREDICATE":
            raise ValueError("'type' should be 'BATCH_PREDICATE'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.batchType.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.batchType.id


class BatchStructureValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr


class BatchStructureParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[BatchStructureValue]

    @validator("type")
    def validate_type(cls, value):
        if value != "BATCH_STRUCTURE":
            raise ValueError("'type' should be 'BATCH_STRUCTURE'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class BatchTagKeyParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[TagKeyValue]

    @validator("type")
    def validate_type(cls, value):
        if value != "BATCH_TAG_KEY":
            raise ValueError("'type' should be 'BATCH_TAG_KEY'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class BatchTagValueParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[TagValueValue]

    @validator("type")
    def validate_type(cls, value):
        if value != "BATCH_TAG_VALUE":
            raise ValueError("'type' should be 'BATCH_TAG_KEY'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class EventTypeValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr


class EventTypeParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[EventTypeValue]

    @validator("type")
    def validate_type(cls, value):
        if value != "EVENT_TYPE":
            raise ValueError("'type' should be 'EVENT_TYPE'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class EventTagKeyValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr


class EventTagKeyParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[EventTagKeyValue]

    @validator("type")
    def validate_type(cls, value):
        if value != "EVENT_TAG_KEY":
            raise ValueError("'type' should be 'EVENT_TAG_KEY'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


class EventTagValueValue(BaseModel):
    type: StrictStr
    id: StrictStr
    reference: StrictStr


class EventTagValueParameter(BaseModel):
    type: StrictStr
    sourceCodeName: StrictStr
    value: Optional[EventTagValueValue]

    @validator("type")
    def validate_type(cls, value):
        if value != "EVENT_TAG_VALUE":
            raise ValueError("'type' should be 'EVENT_TAG_VALUE'")
        return value

    @property
    def readable_value(self):
        if self.value is None:
            return None
        else:
            return self.value.reference

    @property
    def technical_value(self):
        if self.value is None:
            return None
        else:
            return self.value.id


InstanceParameter = Union[
    StringParameter,
    BooleanParameter,
    NumericParameter,
    FileParameter,
    PeriodParameter,
    DurationParameter,
    InstantParameter,
    UnitParameter,
    MeasurementParameter,
    TagKeyParameter,
    TagValueParameter,
    DataParameter,
    StoredContinuousDataParameter,
    StoredBatchDataParameter,
    ComputedContinuousDataParameter,
    ComputedBatchDataParameter,
    BatchTimeDataParameter,
    BatchPredicateParameter,
    BatchStructureParameter,
    BatchTagKeyParameter,
    BatchTagValueParameter,
    EventTypeParameter,
    EventTagKeyParameter,
    EventTagValueParameter,
]


# Model instance
class SingleObservationTimeContext(BaseModel):
    type: StrictStr
    overlappingPeriod: timedelta
    aggregationPeriod: StrictStr

    @validator("type")
    def check_type(cls, value):
        if value != "time":
            raise ValueError("Single observation type should be 'time'")
        return value

    @validator("overlappingPeriod", pre=True)
    def parse_overlapping_period(cls, value):
        return pd.Timedelta(value).to_pytimedelta()


class SingleObservationBatchContext(BaseModel):
    type: StrictStr
    batchPredicate: BatchPredicateValue
    overlappingPeriod: timedelta

    @validator("type")
    def check_type(cls, value):
        if value != "batch":
            raise ValueError("Single observation type should be 'batch'")
        return value

    @validator("overlappingPeriod", pre=True)
    def parse_overlapping_period(cls, value):
        return pd.Timedelta(value).to_pytimedelta()


SingleObservationContext = Union[
    SingleObservationBatchContext, SingleObservationTimeContext
]


class ModelInstance(BaseModel):
    id: StrictStr
    trigger: Trigger
    active: StrictBool
    dataExchangeMode: StrictStr
    singleObservationContext: Optional[SingleObservationContext]
    inputParameters: List[InstanceParameter]
    outputParameters: List[InstanceParameter]


# Model execution
class ModelExecution(BaseModel):
    testMode: StrictBool
    credentials: Optional[Credentials]
    lastSuccessfulExecutionInstant: datetime
    currentExecutionInstant: datetime
    pythonModelInstance: ModelInstance

    def set_as_default_model_execution(self):
        set_default_model_execution(self)

    def find_inputs_by_types(self, input_types: Optional[List[str]] = None):
        if input_types is None:
            return self.pythonModelInstance.inputParameters
        else:
            return [
                param
                for param in self.pythonModelInstance.inputParameters
                if param.type in input_types
            ]

    def find_input_by_sourcecodename(self, source_code_name: str):
        for param in self.pythonModelInstance.inputParameters:
            if param.sourceCodeName == source_code_name:
                return param
            raise KeyError(
                f"No input parameter is referenced by the source code name {source_code_name}"
            )

    def find_outputs_by_types(self, output_types: Optional[List[str]] = None):
        if output_types is None:
            return self.pythonModelInstance.outputParameters
        else:
            return [
                param
                for param in self.pythonModelInstance.outputParameters
                if param.type in output_types
            ]

    def find_output_by_sourcecodename(self, source_code_name: str):
        for param in self.pythonModelInstance.outputParameters:
            if param.sourceCodeName == source_code_name:
                return param
            raise KeyError(
                f"No output parameter is referenced by the source code name {source_code_name}"
            )

    def get_input_dict(
        self,
        input_types: Optional[List[str]] = None,
        mode: str = "object",
        exclude_empty_values: bool = True,
    ):
        input_params = self.find_inputs_by_types(input_types)
        if mode == "reference":
            input_dict = {
                param.sourceCodeName: param.readable_value for param in input_params
            }
        elif mode == "id":
            input_dict = {
                param.sourceCodeName: param.technical_value for param in input_params
            }
        elif mode == "object":
            input_dict = {param.sourceCodeName: param.value for param in input_params}
        else:
            raise ValueError(f"Unknown 'mode': {mode}")

        if exclude_empty_values is True:
            input_dict = {k: v for k, v in input_dict.items() if v is not None}

        return input_dict

    def get_output_dict(
        self,
        output_types: Optional[List[str]] = None,
        mode: str = "object",
        exclude_empty_values: bool = True,
    ):
        output_params = self.find_outputs_by_types(output_types)
        if mode == "reference":
            output_dict = {
                param.sourceCodeName: param.readable_value for param in output_params
            }
        elif mode == "id":
            output_dict = {
                param.sourceCodeName: param.technical_value for param in output_params
            }
        elif mode == "object":
            output_dict = {param.sourceCodeName: param.value for param in output_params}
        else:
            raise ValueError(f"Unknown value for 'mode': {mode}")

        if exclude_empty_values is True:
            output_dict = {k: v for k, v in output_dict.items() if v is not None}

            return output_dict

    def get_data_input_dict(
        self,
        data_type: str = "any",
        strict: bool = False,
        mode: str = "object",
        exclude_empty_values: bool = True,
    ):
        if data_type == "any":
            input_types = [
                "DATA",
                "STORED_CONTINUOUS_DATA",
                "STORED_BATCH_DATA",
                "COMPUTED_CONTINUOUS_DATA",
                "COMPUTED_BATCH_DATA",
                "BATCH_TIME_DATA",
                "BATCH_TAG_KEY",
            ]
        elif data_type == "time":
            input_types = ["STORED_CONTINUOUS_DATA", "COMPUTED_CONTINUOUS_DATA"]
            if strict is False:
                input_types.append("DATA")
        elif data_type == "batch":
            input_types = [
                "STORED_BATCH_DATA",
                "COMPUTED_BATCH_DATA",
                "BATCH_TIME_DATA",
                "BATCH_TAG_KEY",
            ]
            if strict is False:
                input_types.append("DATA")
        else:
            raise ValueError(f"Unknown 'data_type': {data_type}")

        return self.get_input_dict(
            input_types=input_types,
            mode=mode,
            exclude_empty_values=exclude_empty_values,
        )

    def get_data_output_dict(
        self,
        data_type: str = "any",
        strict: bool = False,
        mode: str = "object",
        exclude_empty_values: bool = True,
    ):
        if data_type == "any":
            output_types = [
                "DATA",
                "STORED_CONTINUOUS_DATA",
                "STORED_BATCH_DATA",
                "COMPUTED_CONTINUOUS_DATA",
                "COMPUTED_BATCH_DATA",
                "BATCH_TIME_DATA",
                "BATCH_TAG_KEY",
            ]
        elif data_type == "time":
            output_types = ["STORED_CONTINUOUS_DATA", "COMPUTED_CONTINUOUS_DATA"]
            if strict is False:
                output_types.append("DATA")
        elif data_type == "batch":
            output_types = [
                "STORED_BATCH_DATA",
                "COMPUTED_BATCH_DATA",
                "BATCH_TIME_DATA",
                "BATCH_TAG_KEY",
            ]
            if strict is False:
                output_types.append("DATA")
        else:
            raise ValueError(f"Unknown 'data_type': {data_type}")

        return self.get_output_dict(
            output_types=output_types,
            mode=mode,
            exclude_empty_values=exclude_empty_values,
        )

    @property
    def execution_kwargs(self):
        return {
            "model_instance_id": self.pythonModelInstance.id,
            "test_mode": self.testMode,
            "current_execution_date": self.currentExecutionInstant,
            "last_execution_date": self.lastSuccessfulExecutionInstant,
            "input_parameters": self.get_input_dict(mode="object"),
            "input_parameter_references": self.get_input_dict(mode="reference"),
            "input_parameter_ids": self.get_input_dict(mode="id"),
            "output_parameters": self.get_output_dict(mode="object"),
            "output_parameter_references": self.get_output_dict(mode="reference"),
            "output_parameter_ids": self.get_output_dict(mode="id"),
        }

    def get_execution_argument_value(self, argument_name: str):
        return self.execution_kwargs[argument_name]


# Current event
DEFAULT_EVENT: Optional[ModelExecution] = None


def set_default_model_execution(event: ModelExecution):
    """
    Put an execution event into a global variable, for external use

    Parameters
    ----------
    event: ModelExecution
    """

    global DEFAULT_EVENT
    DEFAULT_EVENT = event


def get_default_model_execution():
    global DEFAULT_EVENT
    return DEFAULT_EVENT



from typing import Optional
import json

from .. import utils
from .. import _dtos
from ... import api

__all__ = ["load_model", "parse_model_execution", "setup_tests"]


def load_model(model_source_path: str):
    return utils.load_module_from_file(model_source_path)


def parse_model_execution(model_execution_str: str):
    return _dtos.ModelExecution(**json.loads(model_execution_str))


def setup_tests(
    credentials: Optional[api.OIAnalyticsAPICredentials] = None,
    model_execution: Optional[_dtos.ModelExecution] = None,
):
    if credentials is not None:
        credentials.set_as_default_credentials()
        print("Provided credentials have been set as default ones")
    else:
        print("No API credentials were provided for setting the default ones")

    if model_execution is not None:
        model_execution.set_as_default_model_execution()
        print("Provided model execution has been set as the default one")
    else:
        print(
            "No model execution configuration was provided for setting the default one"
        )



from typing import Optional
import requests
import io

from ... import api
from .. import _dtos

__all__ = ["get_resource_file"]


def get_resource_file(
    resource_file_id: str,
    model_instance_id: Optional[str] = None,
    api_credentials: Optional[api.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = api.get_default_oianalytics_credentials()

    # Get model execution from environment if not provided
    if model_instance_id is None:
        model_instance_id = _dtos.get_default_model_execution().pythonModelInstance.id

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/python-model-instances/{model_instance_id}/resources/{resource_file_id}/content"
    response = requests.get(
        url=url,
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return io.BytesIO(response.content)



from .batches import (
    get_batch_types,
    get_batch_type_details,
    get_batches,
    update_batch_values,
    update_batch_feature_values,
)
from .data import (
    get_data_list,
    get_time_values,
    get_batch_values,
    get_multiple_data_values,
    insert_time_values,
)
from .events import get_event_types, get_events
from .files import get_file_uploads, read_file_from_file_upload
from .users import get_users

__all__ = [
    "get_batch_types",
    "get_batch_type_details",
    "get_batches",
    "update_batch_values",
    "update_batch_feature_values",
    "get_data_list",
    "get_time_values",
    "get_batch_values",
    "get_multiple_data_values",
    "insert_time_values",
    "get_event_types",
    "get_events",
    "get_file_uploads",
    "read_file_from_file_upload",
    "get_users",
]



from typing import Optional

__all__ = [
    "OIAnalyticsAPICredentials",
    "set_default_oianalytics_credentials",
    "get_default_oianalytics_credentials",
]


# Base class for credentials
class OIAnalyticsAPICredentials:
    def __init__(
        self,
        base_url: str,
        login: Optional[str] = None,
        pwd: Optional[str] = None,
        token: Optional[str] = None,
    ):
        if login is not None and pwd is not None:
            self.auth_type = "Basic"
        elif token is not None:
            self.auth_type = "Token"
        else:
            raise ValueError("Either login/password or token should be provided")
        self.base_url = base_url.strip("/")
        if self.base_url[:5].lower() != "https":
            raise ValueError("Base URL should start with 'https'")
        self.login = login
        self.pwd = pwd
        self.token = token

    def __repr__(self):
        if self.auth_type == "Basic":
            auth_summary = f"Login: {self.login}\nPassword: {self.pwd}"
        elif self.auth_type == "Token":
            auth_summary = f"Token: {self.token}"
        else:
            raise ValueError(
                "The only supported authentication types are Basic or Token"
            )
        return f"{self.auth_type} authentication on {self.base_url}\n{auth_summary}"

    @property
    def auth_kwargs(self):
        if self.auth_type == "Basic":
            return {"auth": (self.login, self.pwd)}
        elif self.auth_type == "Token":
            return {"headers": {"Authorization": f"Bearer {self.token}"}}

    def set_as_default_credentials(self):
        set_default_oianalytics_credentials(credentials=self)


# Init
DEFAULT_CREDENTIALS = None


# Default credentials management
def set_default_oianalytics_credentials(
    credentials: Optional[OIAnalyticsAPICredentials] = None,
    base_url: Optional[str] = None,
    login: Optional[str] = None,
    pwd: Optional[str] = None,
    token: Optional[str] = None,
):
    global DEFAULT_CREDENTIALS

    if isinstance(credentials, OIAnalyticsAPICredentials):
        DEFAULT_CREDENTIALS = credentials
    else:
        DEFAULT_CREDENTIALS = OIAnalyticsAPICredentials(
            base_url=base_url, login=login, pwd=pwd, token=token
        )


def get_default_oianalytics_credentials():
    global DEFAULT_CREDENTIALS
    return DEFAULT_CREDENTIALS



from typing import Optional, Any, Callable

import pandas as pd

from .. import _dtos
from .. import get_single_obs_batch_data, get_single_obs_time_data

__all__ = [
    "get_input_data",
    "test_load_resources",
    "test_process_data",
    "test_full_model",
]


def get_input_data(
    model_execution: Optional[_dtos.ModelExecution] = None,
    rename_data_to_source_code_name: bool = True,
):
    # Get default model_execution if not provided
    if model_execution is None:
        model_execution = _dtos.get_default_model_execution()

    # Get data
    if model_execution.pythonModelInstance.dataExchangeMode != "SINGLE_OBSERVATION":
        raise ValueError(
            "The provided model execution input is not in single observation mode"
        )

    if model_execution.pythonModelInstance.singleObservationContext.type == "time":
        return get_single_obs_time_data(
            model_execution=model_execution,
            rename_data_to_source_code_name=rename_data_to_source_code_name,
        )

    elif model_execution.pythonModelInstance.singleObservationContext.type == "batch":
        return get_single_obs_batch_data(
            model_execution=model_execution,
            rename_data_to_source_code_name=rename_data_to_source_code_name,
        )


def test_load_resources(
    load_resources: Callable,
    model_execution: Optional[_dtos.ModelExecution] = None,
):
    # Get default model_execution if not provided
    if model_execution is None:
        model_execution = _dtos.get_default_model_execution()

    # Run load_resources function
    return load_resources(**model_execution.execution_kwargs)


def test_process_data(
    process_data: Callable,
    data: pd.DataFrame,
    resources: Optional[Any] = None,
    model_execution: Optional[_dtos.ModelExecution] = None,
):
    # Get default model_execution if not provided
    if model_execution is None:
        model_execution = _dtos.get_default_model_execution()

    # Init outputs storage
    outputs = {}

    # Run process_data function
    model_outputs = data.apply(
        lambda r: pd.Series(process_data(resources=resources, **r.to_dict())),
        axis=1,
    )
    outputs["raw_outputs"] = model_outputs

    # Extract configured output data
    output_data = model_outputs[
        [
            col
            for col in model_outputs.columns
            if col in model_execution.get_data_output_dict().keys()
        ]
    ]

    if model_execution.pythonModelInstance.singleObservationContext.type == "batch":
        output_features = model_outputs[
            [
                col
                for col in model_outputs.columns
                if col
                in model_execution.get_output_dict(
                    output_types=["BATCH_TAG_KEY"]
                ).keys()
            ]
        ]

    # Rename outputs
    if model_execution.pythonModelInstance.singleObservationContext.type == "time":
        output_data_renaming_dict = model_execution.get_data_output_dict(
            data_type="any", mode="reference"
        )
    elif model_execution.pythonModelInstance.singleObservationContext.type == "batch":
        output_data_renaming_dict = model_execution.get_data_output_dict(
            data_type="any", mode="id"
        )

    output_data = output_data.rename(columns=output_data_renaming_dict)
    outputs["output_data"] = output_data

    if model_execution.pythonModelInstance.singleObservationContext.type == "batch":
        output_features_renaming_dict = model_execution.get_output_dict(
            output_types=["BATCH_TAG_KEY"], mode="id"
        )

        output_features = output_features.rename(columns=output_features_renaming_dict)
        outputs["output_features"] = output_features

    # Output
    return outputs


def test_full_model(
    load_resources: Callable,
    process_data: Callable,
    model_execution: Optional[_dtos.ModelExecution] = None,
):
    # Load data
    data = get_input_data(model_execution=model_execution)

    # Load resources
    resources = test_load_resources(
        load_resources=load_resources, model_execution=model_execution
    )

    # Process data
    return test_process_data(
        process_data=process_data,
        data=data,
        resources=resources,
        model_execution=model_execution,
    )



from typing import Optional, List, Union
from datetime import datetime, timedelta

import requests
import pandas as pd

from .. import _credentials
from .. import utils

__all__ = [
    "get_data_list",
    "get_data_details",
    "get_time_values",
    "get_batch_values",
    "get_multiple_data_values",
    "insert_time_values",
]


def get_data_list(
    query: Optional[str] = None,
    types: Optional[List[str]] = None,
    measurement_id: Optional[str] = None,
    measurement_name: Optional[str] = None,
    tag_value_id: Optional[List[str]] = None,
    page: Optional[int] = None,
    page_size: Optional[int] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/data"
    response = requests.get(
        url=url,
        params={
            "query": query,
            "types": types,
            "measurement-id": measurement_id,
            "measurement-name": measurement_name,
            "tag-value-id": tag_value_id,
            "page": page,
            "size": page_size,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def get_data_details(
    data_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/data/{data_id}"
    response = requests.get(
        url=url,
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def get_time_values(
    data_id: str,
    start_date: Union[str, datetime],
    end_date: Union[str, datetime],
    aggregation: str,
    number_of_values: Optional[int] = None,
    aggregation_period: Optional[Union[str, timedelta, pd.Timedelta]] = None,
    aggregation_function: Optional[str] = None,
    unit_id: Optional[str] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Format dates
    start_date_iso = utils.get_zulu_isoformat(start_date)
    end_date_iso = utils.get_zulu_isoformat(end_date)

    # Format aggregation period
    aggregation_period_iso = utils.get_iso_period(aggregation_period)

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/data/{data_id}/values"
    response = requests.get(
        url=url,
        params={
            "from": start_date_iso,
            "to": end_date_iso,
            "aggregation": aggregation,
            "number-of-values": number_of_values,
            "aggregation-period": aggregation_period_iso,
            "aggregation-function": aggregation_function,
            "unit-id": unit_id,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def get_batch_values(
    data_id: str,
    start_date: Union[str, datetime],
    end_date: Union[str, datetime],
    aggregation: str,
    number_of_values: Optional[int] = None,
    aggregation_period: Optional[Union[str, timedelta, pd.Timedelta]] = None,
    aggregation_function: Optional[str] = None,
    batch_type_id: Optional[str] = None,
    batch_index_id: Optional[str] = None,
    unit_id: Optional[str] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Format dates
    start_date_iso = utils.get_zulu_isoformat(start_date)
    end_date_iso = utils.get_zulu_isoformat(end_date)

    # Format aggregation period
    aggregation_period_iso = utils.get_iso_period(aggregation_period)

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/data/{data_id}/values"
    response = requests.get(
        url=url,
        params={
            "from": start_date_iso,
            "to": end_date_iso,
            "aggregation": aggregation,
            "number-of-values": number_of_values,
            "aggregation-period": aggregation_period_iso,
            "aggregation-function": aggregation_function,
            "batch-type-id": batch_type_id,
            "batch-index-id": batch_index_id,
            "unit-id": unit_id,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def get_multiple_data_values(
    start_date: Union[str, datetime],
    end_date: Union[str, datetime],
    aggregation: str,
    data_id: Optional[Union[str, List[str]]] = None,
    data_reference: Optional[Union[str, List[str]]] = None,
    number_of_values: Optional[int] = None,
    aggregation_period: Optional[Union[str, timedelta, pd.Timedelta]] = None,
    aggregation_function: Optional[Union[str, List[str]]] = None,
    unit_id: Optional[Union[str, List[str]]] = None,
    unit_label: Optional[Union[str, List[str]]] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Validation
    if not (data_id is None) ^ (data_reference is None):
        raise ValueError(
            "One and only one argument 'data_id' or 'data_reference' should be provided"
        )

    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Format dates
    start_date_iso = utils.get_zulu_isoformat(start_date)
    end_date_iso = utils.get_zulu_isoformat(end_date)

    # Format aggregation period
    aggregation_period_iso = utils.get_iso_period(aggregation_period)

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/data/values"
    response = requests.get(
        url=url,
        params={
            "data-id": data_id,
            "data-reference": data_reference,
            "from": start_date_iso,
            "to": end_date_iso,
            "aggregation": aggregation,
            "number-of-values": number_of_values,
            "aggregation-period": aggregation_period_iso,
            "aggregation-function": aggregation_function,
            "unit-id": unit_id,
            "unit-label": unit_label,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def insert_time_values(
    data: List[dict],
    use_external_reference: bool = False,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/value-upload/time-values"
    response = requests.post(
        url=url,
        json=data,
        params={"use-external-reference": use_external_reference},
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()



import importlib.resources as pkg_resources

from . import _template_resources


# Show templates
def get_template(template: str):
    """
    Retrieve the source of a template

    :param template: str
        Template name, can be either 'free_from_api' or 'single_obs"
    :return: str
        The source code of the template
    """

    return pkg_resources.read_text(_template_resources, f"{template}_template.py")


def print_template(template: str):
    """
    Displays the source of a template

    :param template: str
        Template name, can be either 'free' or 'single_obs"
    :return: str
        The source code of the template
    """

    print(get_template(template=template))



from typing import Optional, List, Union
from datetime import datetime
import io

import requests

from .. import _credentials
from .. import utils

__all__ = [
    "get_file_uploads",
    "get_file_upload_details",
    "get_file_from_file_upload",
    "upload_file",
]


def get_file_uploads(
    start_date: Union[str, datetime],
    end_date: Union[str, datetime],
    filename: Optional[str] = None,
    statuses: Optional[Union[str, List[str]]] = None,
    page: Optional[int] = None,
    page_size: Optional[int] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Format dates
    start_date_iso = utils.get_zulu_isoformat(start_date)
    end_date_iso = utils.get_zulu_isoformat(end_date)

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/file-uploads"
    response = requests.get(
        url=url,
        params={
            "filename": filename,
            "statuses": statuses,
            "from": start_date_iso,
            "to": end_date_iso,
            "page": page,
            "size": page_size,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def get_file_upload_details(
    file_upload_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/file-uploads/{file_upload_id}"
    response = requests.get(
        url=url,
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def get_file_from_file_upload(
    file_upload_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/file-uploads/{file_upload_id}/file"
    response = requests.get(
        url=url,
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.content


def upload_file(
    file_content: Union[io.StringIO, io.BytesIO],
    file_name: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/file-uploads"
    response = requests.post(
        url=url,
        files={"file": (file_name, file_content)},
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()



from typing import Union, Optional, List, Callable
import dateutil
from datetime import datetime, timezone, timedelta
from concurrent.futures import ThreadPoolExecutor
import re

import numpy as np
import pandas as pd

__all__ = [
    "get_zulu_isoformat",
    "get_iso_period",
    "expand_dataframe_column",
    "concat_pages_to_dataframe",
    "Duration",
]


def get_zulu_isoformat(date: Union[str, datetime]):
    # Parse date if it is a string
    if isinstance(date, str):
        date = dateutil.parser.parse(date)

    # Serialize date
    return date.astimezone(timezone.utc).isoformat().replace("+00:00", "Z")


def get_iso_period(period: Optional[Union[str, timedelta, pd.Timedelta]]):
    if period is None:
        return None
    elif isinstance(period, str):
        return period
    elif isinstance(period, (timedelta, pd.Timedelta)):
        return pd.Timedelta(period).isoformat()
    else:
        raise TypeError(f"Can't evaluate iso period from type: {type(period)}")


def to_int(s: Optional[str]):
    return None if s is None else int(s)


class Duration:
    def __init__(
        self,
        years: Optional[int] = None,
        months: Optional[int] = None,
        weeks: Optional[int] = None,
        days: Optional[int] = None,
        hours: Optional[int] = None,
        minutes: Optional[int] = None,
        seconds: Optional[int] = None,
    ):
        self.years = years
        self.months = months
        self.weeks = weeks
        self.days = days
        self.hours = hours
        self.minutes = minutes
        self.seconds = seconds

    @classmethod
    def from_iso(cls, period):
        m = re.match(
            r"^P(?:(?P<years>\d+)Y)?(?:(?P<weeks>\d+)W)?(?:(?P<months>\d+)M)?(?:(?P<days>\d+)D)?T?(?:(?P<hours>\d+)H)?(?:(?P<minutes>\d+)M)?(?:(?P<seconds>\d+(?:.\d+)?)S)?$",
            period,
        )
        return cls(**{k: to_int(v) for k, v in m.groupdict().items()})

    def to_dict(self):
        return {k: v for k, v in self.__dict__.items() if v is not None}

    def to_timedelta(self):
        if self.years is not None or self.months is not None:
            raise ValueError("Pandas Timedelta doesn't accept years and months")

        return pd.Timedelta(**self.to_dict())

    def to_resample_rule(self):
        if list(self.to_dict().keys()) == ["years"]:
            return f"{self.years}YS"
        elif list(self.to_dict().keys()) == ["months"]:
            return f"{self.months}MS"
        elif list(self.to_dict().keys()) == ["weeks"]:
            return f"{self.weeks}W-MON"
        else:
            return self.to_timedelta()


def expand_dataframe_column(
    df: pd.DataFrame,
    col: str,
    add_prefix: bool = True,
    expected_keys: Optional[List[str]] = None,
):
    col_loc = df.columns.get_loc(col)

    # Expand the target column
    if df.shape[0] == 0:
        if expected_keys is None:
            expanded_col = df[col].to_frame()
        else:
            expanded_col = pd.DataFrame(index=df.index, columns=expected_keys)
    else:
        expanded_col = df[col].apply(pd.Series, dtype="object")

    if expanded_col.shape[1] == 0:
        expanded_col[col] = np.NaN

    # Rename generated column using prefix
    if add_prefix is True:
        expanded_col = expanded_col.add_prefix(f"{col}_")

    # Concatenate parts of dataframe
    expanded_df = pd.concat(
        [df[df.columns[:col_loc]], expanded_col, df[df.columns[col_loc + 1 :]]], axis=1
    )

    return expanded_df


def concat_pages_to_dataframe(
    getter: Callable,
    parser: Callable,
    page: Optional[int] = None,
    get_all_pages: bool = True,
    multithread_pages: bool = False,
):
    # Init
    if page is None:
        page = 0

    def get_and_parse_page(page_num: int):
        page_response = getter(page_num)
        return parser(page_response)

    # Get first page
    response = getter(page_num=page)
    df = parser(response)

    total_pages = response["totalPages"]
    if get_all_pages is True and total_pages > 1:
        pages_to_get = range(1, total_pages)

        if multithread_pages is True:
            with ThreadPoolExecutor() as pool:
                page_dfs = list(pool.map(get_and_parse_page, pages_to_get))
        else:
            page_dfs = []
            for i in pages_to_get:
                page_dfs.append(get_and_parse_page(page_num=i))

        # Concatenate additional pages with the first one
        df = pd.concat([df] + page_dfs)

    # Concat dataframes
    return df



from typing import Optional

import pandas as pd

from .. import _credentials
from .. import endpoints
from .. import utils

__all__ = ["get_users"]


def get_users(
    page: int = 0,
    get_all_pages: bool = True,
    multithread_pages: bool = True,
    expand_profile: bool = True,
    extract_from_access_list: Optional[str] = "synthesis",
    expand_access_list: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """
    Get configured users from the OIAnalytics API

    Parameters
    ----------
    page: int, optional
        Page number to retrieve. If None, the first page will be retrieved.
        The argument is ignored if 'get_all_pages' is True.
    get_all_pages: bool, default True
        If True, paging is ignored and all elements are retrieved.
    multithread_pages: bool, default False
        Only used when getting all pages. If True, pages are retrieved in multiple threads simultaneously.
    expand_profile: bool, default True
        Whether or not the profile information should be expanded into multiple columns.
    extract_from_access_list: {'all_access', 'tag_list', None}, default 'synthesis'
        What field should be extracted from access list information.
        'tag_list' will only extract the listed accessible tag keys.
        'all_access' will only extract whether or not the user has access to evefything.
        'synthesis' will combine the previous 2 extracting a list of accessible tag keys or 'All' for a full access.
        If None, the full dictionary is kept.
    expand_access_list: bool, default True
        Whether or not the access list information should be expanded into multiple columns.
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    A DataFrame listing users and their access information
    """

    # Args validation
    if extract_from_access_list not in ["all_access", "tag_list", "synthesis", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_access_list': {extract_from_access_list}"
        )

    if extract_from_access_list is None and expand_access_list is True:
        raise ValueError(
            "Access list cannot be expanded if 'extract_from_access_list' is None"
        )

    # Init
    if get_all_pages is True:
        page = 0

    def get_page(page_num: int):
        page_response = endpoints.users.get_users(page=page_num, api_credentials=api_credentials)
        return page_response

    def parse_page(page_response: dict):
        page_df = pd.DataFrame(page_response["content"])
        page_df["expirationDate"] = pd.to_datetime(page_df["expirationDate"])
        if expand_profile is True:
            page_df = utils.expand_dataframe_column(page_df, "profile")

        # Format acess list
        if extract_from_access_list == "tag_list":
            page_df["accessList"] = page_df["accessList"].apply(
                lambda al: {a["tagKey"]: a["accessibleTagValues"] for a in al}
            )
        elif extract_from_access_list == "all_access":
            page_df["accessList"] = page_df["accessList"].apply(
                lambda al: {a["tagKey"]: a["allTagValuesAccessible"] for a in al}
            )
        elif extract_from_access_list == "synthesis":
            page_df["accessList"] = page_df["accessList"].apply(
                lambda al: {
                    a["tagKey"]: "All"
                    if a["allTagValuesAccessible"] is True
                    else a["accessibleTagValues"]
                    for a in al
                }
            )

        if expand_access_list is True and extract_from_access_list is not None:
            page_df = utils.expand_dataframe_column(
                page_df, "accessList", add_prefix=False
            )

        page_df.set_index("id", inplace=True)
        return page_df

    # Query endpoint
    df = utils.concat_pages_to_dataframe(
        getter=get_page,
        parser=parse_page,
        page=page,
        get_all_pages=get_all_pages,
        multithread_pages=multithread_pages,
    )

    # Output
    return df



from typing import Optional, Any, Callable

from .. import _dtos
from .. import outputs

__all__ = [
    "test_load_data",
    "test_load_resources",
    "test_process_data",
    "test_full_model",
]


def test_load_data(
    load_data: Callable,
    model_execution: Optional[_dtos.ModelExecution] = None,
):
    # Get default model_execution if not provided
    if model_execution is None:
        model_execution = _dtos.get_default_model_execution()

    # Run load_data function
    return load_data(**model_execution.execution_kwargs)


def test_load_resources(
    load_resources: Callable,
    model_execution: Optional[_dtos.ModelExecution] = None,
):
    # Get default model_execution if not provided
    if model_execution is None:
        model_execution = _dtos.get_default_model_execution()

    # Run load_resources function
    return load_resources(**model_execution.execution_kwargs)


def test_process_data(
    process_data: Callable,
    data: Optional[Any] = None,
    resources: Optional[Any] = None,
    model_execution: Optional[_dtos.ModelExecution] = None,
):
    # Get default model_execution if not provided
    if model_execution is None:
        model_execution = _dtos.get_default_model_execution()

    # Run process_data function
    model_output = process_data(
        data=data, resources=resources, **model_execution.execution_kwargs
    )

    if not isinstance(
        model_output,
        (
            outputs.OIModelOutputs,
            outputs.FileOutput,
            outputs.TimeValuesOutput,
            outputs.Delay,
        ),
    ):
        print(
            "Warning: the model output should be an instance of an output class from oianalytics.models.outputs"
        )

    return model_output


def test_full_model(
    load_data: Callable,
    load_resources: Callable,
    process_data: Callable,
    model_execution: Optional[_dtos.ModelExecution] = None,
):
    # Get default model_execution if not provided
    if model_execution is None:
        model_execution = _dtos.get_default_model_execution()

    # Run load_data function
    data = load_data(**model_execution.execution_kwargs)

    # Run load_resources function
    resources = load_resources(**model_execution.execution_kwargs)

    # Run load_data function
    model_output = process_data(
        data=data, resources=resources, **model_execution.execution_kwargs
    )

    if not isinstance(
        model_output,
        (
            outputs.OIModelOutputs,
            outputs.FileOutput,
            outputs.TimeValuesOutput,
            outputs.Delay,
        ),
    ):
        print(
            "Warning: the model output should be an instance of an output class from oianalytics.models.outputs"
        )

    return model_output



from typing import Union, Optional
import io
import time

import pandas as pd

from .. import api
from ._dtos import get_default_model_execution

__all__ = [
    "FileOutput",
    "TimeValuesOutput",
    "Delay",
    "BatchValuesOutput",
    "BatchFeaturesOutput",
    "OIModelOutputs",
]


# Output classes
class FileOutput:
    def __init__(self, file_name: str, content: Union[io.StringIO, io.BytesIO]):
        self.output_type = "file"
        self.file_name = file_name
        self.content = content

    @classmethod
    def from_pandas(
        cls,
        data: Union[pd.Series, pd.DataFrame],
        file_name: str,
        file_type: str = "csv",
        writing_kwargs: Optional[dict] = None,
    ):
        # Init
        if writing_kwargs is None:
            writing_kwargs = {}

        bio = io.BytesIO()

        # Write data
        if file_type == "excel":
            data.to_excel(bio, **writing_kwargs)
        elif file_type == "csv":
            data.to_csv(bio, **writing_kwargs)
        else:
            raise NotImplementedError(f"Unsupported file_type: {file_type}")
        bio.seek(0)

        # Create object
        return cls(file_name=file_name, content=bio)

    def send_to_oianalytics(
        self, api_credentials: Optional[api.OIAnalyticsAPICredentials] = None
    ):
        return api.endpoints.files.upload_file(
            file_content=self.content,
            file_name=self.file_name,
            api_credentials=api_credentials,
        )


class TimeValuesOutput:
    def __init__(
        self,
        data: Union[pd.Series, pd.DataFrame],
        units: Optional[dict] = None,
        rename_data: bool = True,
        use_external_reference: bool = False,
        timestamp_index_name: str = "timestamp",
    ):
        self.output_type = "time_values"

        # Rename data if specified
        data_df = data.to_frame() if isinstance(data, pd.Series) else data

        if rename_data is True:
            model_exec = get_default_model_execution()
            if model_exec is None:
                raise ValueError(
                    "Data can't be renamed without a current model_exec set globally"
                )

            output_dict = model_exec.get_data_output_dict(
                data_type="time", mode="reference"
            )
            self.data = data_df.rename(columns=output_dict)
        else:
            self.data = data_df

        # Specify units
        if units is None:
            model_exec = get_default_model_execution()
            if model_exec is not None:
                self.units = {
                    output_data.reference: output_data.unit.label
                    for output_data in model_exec.get_data_output_dict(
                        data_type="time", mode="object"
                    ).values()
                }
        else:
            self.units = units

        self.use_external_reference = use_external_reference

        self.timestamp_index_name = timestamp_index_name

    def send_to_oianalytics(
        self, api_credentials: Optional[api.OIAnalyticsAPICredentials] = None
    ):
        return api.insert_time_values(
            data=self.data,
            units=self.units,
            use_external_reference=self.use_external_reference,
            timestamp_index_name=self.timestamp_index_name,
            api_credentials=api_credentials,
        )


class BatchValuesOutput:
    def __init__(
        self,
        batch_type_id: str,
        data: Union[pd.Series, pd.DataFrame],
        units: Optional[dict] = None,
        batch_id_index_name: str = "batch_id",
        rename_data: bool = True,
    ):
        self.output_type = "batch_values"
        self.batch_type_id = batch_type_id

        # Rename data if specified
        data_df = data.to_frame() if isinstance(data, pd.Series) else data

        if rename_data is True:
            model_exec = get_default_model_execution()
            if model_exec is None:
                raise ValueError(
                    "Data can't be renamed without a current model_exec set globally"
                )

            output_dict = model_exec.get_data_output_dict(data_type="batch", mode="id")
            self.data = data_df.rename(columns=output_dict)
        else:
            self.data = data_df

        # Specify units
        if units is None:
            model_exec = get_default_model_execution()
            if model_exec is not None:
                self.units = {
                    output_data.reference: output_data.unit.id
                    for output_data in model_exec.get_data_output_dict(
                        data_type="batch", mode="object"
                    ).values()
                }
        else:
            self.units = units

        self.batch_id_index_name = batch_id_index_name

    def send_to_oianalytics(
        self, api_credentials: Optional[api.OIAnalyticsAPICredentials] = None
    ):
        return api.update_batch_values(
            batch_type_id=self.batch_type_id,
            data=self.data,
            unit_ids=self.units,
            batch_id_index_name=self.batch_id_index_name,
            api_credentials=api_credentials,
        )


class BatchFeaturesOutput:
    def __init__(
        self,
        batch_type_id: str,
        data: Union[pd.Series, pd.DataFrame],
        rename_features: bool = True,
        batch_id_index_name: str = "batch_id",
    ):
        self.output_type = "batch_features"
        self.batch_type_id = batch_type_id

        # Rename data if specified
        data_df = data.to_frame() if isinstance(data, pd.Series) else data

        if rename_features is True:
            model_exec = get_default_model_execution()
            if model_exec is None:
                raise ValueError(
                    "Features can't be renamed without a current model_exec set globally"
                )

            output_dict = model_exec.get_data_output_dict(data_type="batch", mode="id")
            self.data = data_df.rename(columns=output_dict)
        else:
            self.data = data_df

        self.batch_id_index_name = batch_id_index_name

    def send_to_oianalytics(
        self, api_credentials: Optional[api.OIAnalyticsAPICredentials] = None
    ):
        return api.update_batch_feature_values(
            batch_type_id=self.batch_type_id,
            data=self.data,
            batch_id_index_name=self.batch_id_index_name,
            api_credentials=api_credentials,
        )


class Delay:
    def __init__(self, duration=5):
        self.output_type = "delay"
        self.duration = duration

    def send_to_oianalytics(
        self, api_credentials: Optional[api.OIAnalyticsAPICredentials] = None
    ):
        time.sleep(self.duration)


class OIModelOutputs:
    def __init__(self):
        self.output_type = "outputs_queue"
        self.model_outputs = []

    def add_output(self, output_object: Union[FileOutput, TimeValuesOutput, Delay]):
        self.model_outputs.append(output_object)

    def send_to_oianalytics(
        self, api_credentials: Optional[api.OIAnalyticsAPICredentials] = None
    ):
        for model_output in self.model_outputs:
            model_output.send_to_oianalytics(api_credentials=api_credentials)



import importlib.machinery


def reverse_dict(d: dict):
    return {v: k for k, v in d.items()}


def load_module_from_file(path: str):
    loader = importlib.machinery.SourceFileLoader("loaded_module", path)
    return loader.load_module("loaded_module")



from typing import Optional

import requests

from .. import _credentials

__all__ = ["get_users", "get_user_details"]


def get_users(
    page: Optional[int] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/users"
    response = requests.get(
        url=url,
        params={"page": page},
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def get_user_details(
    user_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/users/{user_id}"
    response = requests.get(
        url=url,
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['oianalytics',
 'oianalytics.api',
 'oianalytics.api._dataframes',
 'oianalytics.api.endpoints',
 'oianalytics.models',
 'oianalytics.models._queries',
 'oianalytics.models._template_resources',
 'oianalytics.models.testing']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=1.3.4,<2.0.0', 'pydantic>=1.9.0,<2.0.0', 'requests>=2.26.0,<3.0.0']

setup_kwargs = {
    'name': 'oianalytics',
    'version': '0.3.4',
    'description': 'Python tools for working with OIAnalytics',
    'long_description': None,
    'author': 'Optimistik SAS',
    'author_email': 'arthur.martel@optimistik.fr',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from oianalytics import models

from datetime import datetime


def load_resources(
    model_instance_id: str,
    test_mode: bool,
    current_execution_date: datetime,
    last_execution_date: datetime,
    input_parameters: dict,
    input_parameter_references: dict,
    input_parameter_ids: dict,
    output_parameters: dict,
    output_parameter_references: dict,
    output_parameter_ids: dict,
    **kwargs
):
    # # Arguments can be absent from the function signature as long **kwargs is present
    # # Only useful arguments can be kept
    # # Keeping **kwargs in the signature is strongly recommended for easier future compatibility
    #
    # # Here's an example on how to retrieve an instance resource
    # my_resource = models.get_resource_file(
    #     resource_file_id=input_parameter_ids["my_resource_sourcecodename"]
    # )
    #
    # # The resources to be used in the 'process_data' function have to be returned
    # # It can be any format since the 'process_data' is responsible of its usage
    # return my_resource
    # return {"my_first_resource": my_resource}
    pass


def process_data(
    model_instance_id: str,
    test_mode: bool,
    current_execution_date: datetime,
    last_execution_date: datetime,
    input_parameters: dict,
    input_parameter_references: dict,
    input_parameter_ids: dict,
    output_parameters: dict,
    output_parameter_references: dict,
    output_parameter_ids: dict,
    **kwargs
):
    # Arguments can be absent from the function signature as long **kwargs is present
    # Only useful arguments can be kept
    # Keeping **kwargs in the signature is strongly recommended for easier future compatibility
    #
    # Input data should be provided as arguments, using their source code name
    # The output should be a dictionary using output source code names as keys
    # The basic following example (which converts a temperature from Celsius to Kelvin) is properly structured
    # def process_data(temp_c, **kwargs):
    #     temp_k = temp_c + 273.15
    #     return {"temp_k": temp_k}
    pass



from typing import Optional

from ... import api
from .. import _dtos
from .. import utils

__all__ = ["get_single_obs_time_data", "get_single_obs_batch_data"]


def get_single_obs_time_data(
    model_execution: Optional[_dtos.ModelExecution] = None,
    rename_data_to_source_code_name: bool = True,
    api_credentials: Optional[api.OIAnalyticsAPICredentials] = None,
):
    # Get model execution from environment if not specified
    if model_execution is None:
        model_execution = _dtos.get_default_model_execution()
        if model_execution.pythonModelInstance.dataExchangeMode != "SINGLE_OBSERVATION":
            raise ValueError("The execution has to be in single observation mode")
        if model_execution.pythonModelInstance.singleObservationContext.type != "time":
            raise ValueError(
                "The execution has to be in single observation mode on time data"
            )

    # Start date
    start_date = (
        model_execution.lastSuccessfulExecutionInstant
        - model_execution.pythonModelInstance.singleObservationContext.overlappingPeriod
    )

    # End date
    end_date = model_execution.currentExecutionInstant

    # Aggregation
    if (
        model_execution.pythonModelInstance.singleObservationContext.aggregationPeriod
        == "PT0S"
    ):
        aggregation = "RAW_VALUES"
    else:
        aggregation = "TIME"

    # Aggregation period
    if aggregation == "RAW_VALUES":
        aggregation_period = None
    else:
        aggregation_period = (
            model_execution.pythonModelInstance.singleObservationContext.aggregationPeriod
        )

    # Aggregation function
    if aggregation == "RAW_VALUES":
        aggregation_function = None
    else:
        aggregation_function = [
            input_data.aggregationFunction
            for input_data in model_execution.get_data_input_dict(
                data_type="any", mode="object"
            ).values()
        ]

    # Get data list
    input_data_references = model_execution.get_data_input_dict(
        data_type="any", mode="reference"
    )
    input_renaming_dict = utils.reverse_dict(input_data_references)

    # Get units
    input_unit_ids = [
        input_data.unit.id
        for input_data in model_execution.get_data_input_dict(
            data_type="any", mode="object"
        ).values()
    ]

    # Query endpoint
    data = api.get_multiple_data_values(
        start_date=start_date,
        end_date=end_date,
        aggregation=aggregation,
        data_id=None,
        data_reference=input_data_references.values(),
        number_of_values=None,
        aggregation_period=aggregation_period,
        aggregation_function=aggregation_function,
        unit_id=input_unit_ids,
        unit_label=None,
        name_data_from="reference",
        append_unit_to_description=False,
        join_series_on="timestamp",
        api_credentials=api_credentials,
    )

    # Rename data
    if rename_data_to_source_code_name is True:
        data = data.rename(columns=input_renaming_dict)

    # Output
    return data


def get_single_obs_batch_data(
    model_execution: Optional[_dtos.ModelExecution] = None,
    rename_data_to_source_code_name: bool = True,
    api_credentials: Optional[api.OIAnalyticsAPICredentials] = None,
):
    # Get variables from environment if not specified
    if model_execution is None:
        model_execution = _dtos.get_default_model_execution()
        if model_execution.pythonModelInstance.dataExchangeMode != "SINGLE_OBSERVATION":
            raise ValueError("The execution has to be in single observation mode")
        if model_execution.pythonModelInstance.singleObservationContext.type != "batch":
            raise ValueError(
                "The execution has to be in single observation mode on batches"
            )

    # Start date
    start_date = (
        model_execution.lastSuccessfulExecutionInstant
        - model_execution.pythonModelInstance.singleObservationContext.overlappingPeriod
    )

    # End date
    end_date = model_execution.currentExecutionInstant

    # Batch type
    batch_type_id = (
        model_execution.pythonModelInstance.singleObservationContext.batchPredicate.batchType.id
    )

    # Batch feature filters
    features_value_ids = [
        feature.id
        for feature in model_execution.pythonModelInstance.singleObservationContext.batchPredicate.featureFilters
    ]

    # Get data list
    input_data_references = model_execution.get_data_input_dict(
        data_type="any", mode="reference"
    )
    input_renaming_dict = utils.reverse_dict(input_data_references)

    # Get units
    input_unit_ids = [
        input_data.unit.id
        for input_data in model_execution.get_data_input_dict(
            data_type="any", mode="object"
        ).values()
    ]

    # Query endpoint for data
    data = None

    if len(model_execution.get_data_input_dict()) > 0:
        data = api.get_multiple_data_values(
            start_date=start_date,
            end_date=end_date,
            aggregation="RAW_VALUES",
            data_reference=input_data_references.values(),
            unit_id=input_unit_ids,
            name_data_from="reference",
            append_unit_to_description=False,
            join_series_on="index",
            api_credentials=api_credentials,
        )

        # Rename data
        if rename_data_to_source_code_name is True:
            data = data.rename(columns=input_renaming_dict)

    # Query endpoint for features & batch predicate
    features = api.get_batches(
        batch_type_id=batch_type_id,
        start_date=start_date,
        end_date=end_date,
        features_value_ids=features_value_ids,
        api_credentials=api_credentials,
    )[1]

    # Rename features
    features_renaming_dict = utils.reverse_dict(
        model_execution.get_input_dict(input_types=["BATCH_TAG_KEY"], mode="reference")
    )

    features = features[
        [col for col in features.columns if col in features_renaming_dict.keys()]
    ]

    if rename_data_to_source_code_name is True:
        features = features.rename(columns=features_renaming_dict)

    if data is None:
        data = features
    else:
        data = data.join(features, how="right")

    # Output
    return data



from oianalytics import api, models

from datetime import datetime


def load_data(
    model_instance_id: str,
    test_mode: bool,
    current_execution_date: datetime,
    last_execution_date: datetime,
    input_parameters: dict,
    input_parameter_references: dict,
    input_parameter_ids: dict,
    output_parameters: dict,
    output_parameter_references: dict,
    output_parameter_ids: dict,
    **kwargs,
):
    # # Arguments can be absent from the function signature as long **kwargs is present
    # # Only useful arguments can be kept
    # # Keeping **kwargs in the signature is strongly recommended for easier future compatibility
    #
    # # Here's an example on how to retrieve a single time data values
    # time_data = api.get_time_values(
    #     data_id=input_parameter_ids["my_timedata_sourcecodename"],
    #     start_date=last_execution_date,
    #     end_date=current_execution_date,
    #     aggregation="RAW_VALUES",
    # )
    #
    # # The data to be used in the 'process_data' function has to be returned
    # # It can be any format since the 'process_data' is responsible of its usage
    # return time_data
    # return {"my_time_data": time_data}
    pass


def load_resources(
    model_instance_id: str,
    test_mode: bool,
    current_execution_date: datetime,
    last_execution_date: datetime,
    input_parameters: dict,
    input_parameter_references: dict,
    input_parameter_ids: dict,
    output_parameters: dict,
    output_parameter_references: dict,
    output_parameter_ids: dict,
    **kwargs,
):
    # # Arguments can be absent from the function signature as long **kwargs is present
    # # Only useful arguments can be kept
    # # Keeping **kwargs in the signature is strongly recommended for easier future compatibility
    #
    # # Here's an example on how to retrieve an instance resource
    # my_resource = models.get_resource_file(
    #     resource_file_id=input_parameter_ids["my_resource_sourcecodename"]
    # )
    #
    # # The resources to be used in the 'process_data' function have to be returned
    # # It can be any format since the 'process_data' is responsible of its usage
    # return my_resource
    # return {"my_first_resource": my_resource}
    pass


def process_data(
    data,
    resources,
    model_instance_id: str,
    test_mode: bool,
    current_execution_date: datetime,
    last_execution_date: datetime,
    input_parameters: dict,
    input_parameter_references: dict,
    input_parameter_ids: dict,
    output_parameters: dict,
    output_parameter_references: dict,
    output_parameter_ids: dict,
    **kwargs,
):
    # # Arguments can be absent from the function signature as long **kwargs is present
    # # Only useful arguments can be kept
    # # Keeping **kwargs in the signature is strongly recommended for easier future compatibility
    #
    # # The output of 'load_data' is sent through the 'data' argument
    # # The output of 'load_resources' is sent through the 'resources' argument
    #
    # # The output of this function should be an object made for sending model results into OIAnalytics
    # # These are the output objects found in models.outputs
    # # Typically it would be a container for multiple output such as the following
    # outputs = models.outputs.OIModelOutputs()
    #
    # # Specific outputs can be added to such a container
    # outputs.add_output(
    #     models.outputs.FileOutput.from_pandas(
    #         data=data, file_name="my_time_values.csv", writing_kwargs={"sep": ";"}
    #     )
    # )
    #
    # # This object has then to be returned by the model
    # # The OIAnalytics application is responsible for sending and actually storing values
    # return outputs
    pass



from typing import Optional, List, Union
from datetime import datetime

import requests

from .. import _credentials
from .. import utils

__all__ = [
    "get_batch_types",
    "get_batch_type_details",
    "get_batches",
    "update_batch_value",
    "update_batch_feature_value",
]


def get_batch_types(
    page: Optional[int] = None,
    page_size: Optional[int] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/batch-types"
    response = requests.get(
        url=url,
        params={
            "page": page,
            "size": page_size,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def get_batch_type_details(
    batch_type_id: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/batch-types/{batch_type_id}"
    response = requests.get(url=url, **api_credentials.auth_kwargs)

    # Output
    response.raise_for_status()
    return response.json()


def get_batches(
    batch_type_id: str,
    start_date: Union[str, datetime],
    end_date: Union[str, datetime],
    name: Optional[str] = None,
    features_value_ids: Optional[Union[str, List[str]]] = None,
    page: Optional[int] = None,
    page_size: Optional[int] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Format dates
    start_date_iso = utils.get_zulu_isoformat(start_date)
    end_date_iso = utils.get_zulu_isoformat(end_date)

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/batch-types/{batch_type_id}/batches"
    response = requests.get(
        url=url,
        params={
            "start": start_date_iso,
            "end": end_date_iso,
            "name": name,
            "feature-values": features_value_ids,
            "page": page,
            "size": page_size,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.json()


def update_batch_value(
    batch_type_id: str,
    batch_id: str,
    data_id: str,
    value: Union[int, float],
    unit_id: Optional[str] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/batch-types/{batch_type_id}/batches/{batch_id}/value"
    response = requests.put(
        url=url,
        json={
            "dataId": data_id,
            "value": value,
            "unitId": unit_id,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.status_code


def update_batch_feature_value(
    batch_type_id: str,
    batch_id: str,
    feature_id: str,
    value: str,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    # Get credentials from environment if not provided
    if api_credentials is None:
        api_credentials = _credentials.get_default_oianalytics_credentials()

    # Query endpoint
    url = f"{api_credentials.base_url}/api/oianalytics/batch-types/{batch_type_id}/batches/{batch_id}/feature"
    response = requests.put(
        url=url,
        json={
            "batchTagKeyId": feature_id,
            "batchTagValueValue": value,
        },
        **api_credentials.auth_kwargs,
    )

    # Output
    response.raise_for_status()
    return response.status_code



from typing import Optional, Union, List, Callable
import io
from datetime import datetime

import pandas as pd

from .. import _credentials
from .. import endpoints
from .. import utils

__all__ = ["get_file_uploads", "read_file_from_file_upload"]


def get_file_uploads(
    start_date: Union[str, datetime],
    end_date: Union[str, datetime],
    filename: Optional[str] = None,
    statuses: Optional[Union[str, List[str]]] = None,
    page: Optional[int] = None,
    page_size: Optional[int] = None,
    get_all_pages: bool = True,
    multithread_pages: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """
    Get a list of file uploads from the OIAnalytics API

    Parameters
    ----------
    start_date: datetime or str
        The beginning of the period to be retrieved
    end_date: datetime or str
        The end of the period to be retrieved
    filename: str, optional
        A portion of text the filename uploaded should contain.
    statuses: str or list of str, optional
        Statuses to filter the file uploads.
        Available statuses are 'PENDING', 'NO_PARSER_FOUND', 'RUNNING', 'DEACTIVATED_PARSER_FOUND', 'SUCCESS', 'ERROR'.
    page: int, optional
        Page number to retrieve. If None, the first page will be retrieved.
        The argument is ignored if 'get_all_pages' is True.
    page_size: int, optional
        The size of each page to retrieve. By default, 20 elements are retrieved.
        The argument is ignored if 'get_all_pages' is True.
    get_all_pages: bool, default True
        If True, paging is ignored and all elements are retrieved.
    multithread_pages: bool, default False
        Only used when getting all pages. If True, pages are retrieved in multiple threads simultaneously.
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    A DataFrame listing file uploads
    """

    # Init
    if get_all_pages is True:
        page = 0
        page_size = 500

    def get_page(page_num: int):
        page_response = endpoints.files.get_file_uploads(start_date=start_date, end_date=end_date, filename=filename,
                                                         statuses=statuses, page=page_num, page_size=page_size,
                                                         api_credentials=api_credentials)
        return page_response

    def parse_page(page_response: dict):
        page_df = pd.DataFrame(page_response["content"])

        # Expected columns if content is empty
        if page_df.shape[0] == 0:
            page_df = pd.DataFrame(
                columns=[
                    "id",
                    "creationInstant",
                    "startInstant",
                    "endInstant",
                    "status",
                    "uploader",
                    "filename",
                ]
            )

        for col in ["creationInstant", "startInstant", "endInstant"]:
            page_df[col] = pd.to_datetime(page_df[col])
        page_df.set_index("id", inplace=True)
        return page_df

    # Query endpoint
    df = utils.concat_pages_to_dataframe(
        getter=get_page,
        parser=parse_page,
        page=page,
        get_all_pages=get_all_pages,
        multithread_pages=multithread_pages,
    )

    # Output
    return df


def read_file_from_file_upload(
    file_upload_id: str,
    parser: Callable = pd.read_csv,
    parser_kwargs: Optional[dict] = None,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """
    Get the content of a file from previous file uploads and read it

    Parameters
    ----------
    file_upload_id: str
        The id of the file to be retrieved
    parser: callable, default pd.read_csv
        A function which is applied to the file content BytesIO
    parser_kwargs: dict, optional
        A dictionary containing kwargs to be used in the parser function when reading the file
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    The result of the file content parsing built as following: parser(file_content, **parser_kwargs)
    """

    # Init
    if parser_kwargs is None:
        parser_kwargs = {}

    # Query endpoint
    response = endpoints.files.get_file_from_file_upload(file_upload_id=file_upload_id, api_credentials=api_credentials)

    # Output
    return parser(io.BytesIO(response), **parser_kwargs)



from typing import Optional, Union, List
from datetime import datetime, timedelta

import pandas as pd

from .. import _credentials
from .. import endpoints
from .. import utils

__all__ = [
    "get_data_list",
    "get_time_values",
    "get_batch_values",
    "get_multiple_data_values",
    "insert_time_values",
]


def get_data_list(
    query: Optional[str] = None,
    types: Optional[List[str]] = None,
    measurement_id: Optional[str] = None,
    measurement_name: Optional[str] = None,
    tag_value_id: Optional[List[str]] = None,
    page: Optional[int] = None,
    page_size: Optional[int] = None,
    get_all_pages: bool = True,
    multithread_pages: bool = True,
    expand_measurement: bool = True,
    extract_from_storage_unit: Optional[str] = "label",
    extract_from_tags: Optional[str] = "value",
    expand_tags: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """
    Get the configured data from the OIAnalytics API

    Parameters
    ----------
    query: str, optional
        A text to search for specific data
    types: list of str, optional
        An array of types to search for specific data types. If omitted then all types are considered.
    measurement_id: str, optional
        The measurement id all data should have. Cannot be used in conjunction with 'measurement_name'.
    measurement_name: str, optional
        The measurement name all data should have. Cannot be used in conjunction with 'measurement_id'.
    tag_value_id: list of str, optional
        An array of tag value ids the data should have.
    page: int, optional
        Page number to retrieve. If None, the first page will be retrieved.
        The argument is ignored if 'get_all_pages' is True.
    page_size: int, optional
        The size of each page to retrieve. By default, 20 elements are retrieved.
        The argument is ignored if 'get_all_pages' is True.
    get_all_pages: bool, default True
        If True, paging is ignored and all elements are retrieved.
    multithread_pages: bool, default False
        Only used when getting all pages. If True, pages are retrieved in multiple threads simultaneously.
    expand_measurement: bool, default True
        Whether or not measurement information should be expanded into multiple columns.
    extract_from_storage_unit: {'id', 'label', None}, default 'label'
        What field should be extracted from storage unit information. If None, the full dictionary is kept.
    extract_from_tags: {'id', 'value', None}, default 'value'
        What field should be extracted for naming tags. If None, the full dictionary is kept.
    expand_tags: bool, default True
        Whether or not tags should be expanded into multiple columns.
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    A DataFrame listing data and their properties
    """

    # Args validation
    if extract_from_storage_unit not in ["id", "label", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_storage_unit': {extract_from_storage_unit}"
        )

    if extract_from_tags not in ["id", "value", None]:
        raise ValueError(
            f"Unexpected value for 'extract_from_tags': {extract_from_tags}"
        )

    # Init
    if get_all_pages is True:
        page = 0
        page_size = 500

    def get_page(page_num: int):
        page_response = endpoints.data.get_data_list(
            query=query,
            types=types,
            measurement_id=measurement_id,
            measurement_name=measurement_name,
            tag_value_id=tag_value_id,
            page=page_num,
            page_size=page_size,
            api_credentials=api_credentials,
        )
        return page_response

    def parse_page(page_response: dict):
        page_df = pd.DataFrame(page_response["content"])

        # Expected columns if content is empty
        if page_df.shape[0] == 0:
            page_df = pd.DataFrame(
                columns=[
                    "dataType",
                    "id",
                    "reference",
                    "description",
                    "measurement",
                    "tags",
                ]
            )

        # Format dataframe
        if expand_measurement is True:
            page_df = utils.expand_dataframe_column(
                page_df,
                "measurement",
                expected_keys=[
                    "id",
                    "name",
                    "storageUnit",
                    "defaultUnitFamily",
                    "quantityName",
                ],
            )

            if extract_from_storage_unit == "id":
                page_df["measurement_storageUnit"] = page_df[
                    "measurement_storageUnit"
                ].apply(lambda su: su["id"])
            elif extract_from_storage_unit == "label":
                page_df["measurement_storageUnit"] = page_df[
                    "measurement_storageUnit"
                ].apply(lambda su: su["label"])

        # Extract from tags
        if extract_from_tags == "id":
            page_df["tags"] = page_df["tags"].apply(
                lambda tl: {t["tagKey"]["id"]: t["id"] for t in tl}
            )
        elif extract_from_tags == "value":
            page_df["tags"] = page_df["tags"].apply(
                lambda tl: {t["tagKey"]["key"]: t["value"] for t in tl}
            )

        if expand_tags is True and extract_from_tags is not None:
            page_df = utils.expand_dataframe_column(page_df, "tags", add_prefix=False)

        page_df.set_index("id", inplace=True)
        return page_df

    # Query endpoint
    df = utils.concat_pages_to_dataframe(
        getter=get_page,
        parser=parse_page,
        page=page,
        get_all_pages=get_all_pages,
        multithread_pages=multithread_pages,
    )

    # Output
    return df


def get_time_values(
    data_id: str,
    start_date: Union[str, datetime],
    end_date: Union[str, datetime],
    aggregation: str,
    number_of_values: Optional[int] = None,
    aggregation_period: Optional[str] = None,
    aggregation_function: Optional[str] = None,
    unit_id: Optional[str] = None,
    name_data_from: str = "reference",
    append_unit_to_description: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """
    Get values from a temporal data from the OIAnalytics API

    Parameters
    ----------
    data_id: str
        The data id on which values are to be retrieved
    start_date: datetime or str
        The beginning of the period to be retrieved
    end_date: datetime or str
        The end of the period to be retrieved
    aggregation: {'RAW_VALUES', 'TIME'}
        How to aggregate the values. If 'TIME', aggregation period and function should be specified.
    number_of_values: int, optional
        If > 0 returns only the number of values specified. Only works for stored data.
    aggregation_period: str, optional
        Required in case 'aggregation' is 'TIME' and is the sampling period of the expected result.
        This period should be an ISO period as described in https://en.wikipedia.org/wiki/ISO_8601#Durations
    aggregation_function: str, optional
        Required in case 'aggregation' is 'TIME' and is the aggregation function to use to aggregate the values within
        the sampling period.
        Possible values are 'FIRST', 'LAST', 'LAST_MINUS_FIRST', 'SUM', 'MIN', 'MAX', 'MEAN', 'MEDIAN', 'STDEV',
        'PERCENTILE5', 'PERCENTILE95', 'DECILE1', 'DECILE9', 'QUARTILE1', 'QUARTILE9', 'COUNT', 'MEAN_MINUS_SIGMA',
        'MEAN_PLUS_SIGMA', 'MEAN_MINUS_TWO_SIGMA', 'MEAN_PLUS_TWO_SIGMA', 'MEAN_MINUS_THREE_SIGMA',
        'MEAN_PLUS_THREE_SIGMA', 'VALUE_CHANGE'.
    unit_id: str, optional
        The id of the unit to use to express the values. If not present a default unit will be used.
        This unit should be compatible with the physical quantity of the data queried.
    name_data_from: {'id', 'reference', 'description'}
        What field should be extracted for naming data.
    append_unit_to_description: bool, default True
        Only used when 'name_data_from' is 'description'. If True, the unit is added after the description.
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    A Series containing the values of the data, indexed by 'timestamp'
    """

    # Args validation
    if name_data_from not in ["id", "reference", "description"]:
        raise ValueError(f"Unexpected value for 'name_data_from': {name_data_from}")

    # Query endpoint
    data = endpoints.data.get_time_values(
        data_id=data_id,
        start_date=start_date,
        end_date=end_date,
        aggregation=aggregation,
        number_of_values=number_of_values,
        aggregation_period=aggregation_period,
        aggregation_function=aggregation_function,
        unit_id=unit_id,
        api_credentials=api_credentials,
    )

    # Format series
    if name_data_from == "id":
        data_name = data["data"]["id"]
    elif name_data_from == "reference":
        data_name = data["data"]["reference"]
    elif name_data_from == "description":
        data_name = data["data"]["description"]
        if append_unit_to_description is True:
            data_name = f'{data_name} ({data["unit"]["label"]})'
    else:
        raise ValueError(f"Unexpected value for 'name_data_from': {name_data_from}")

    if len(data["values"]) == 0:
        timestamp_index = pd.to_datetime(data["timestamps"]).tz_localize("UTC")
    else:
        timestamp_index = pd.to_datetime(data["timestamps"])

    ser = pd.Series(
        name=data_name,
        index=timestamp_index,
        data=data["values"],
        dtype=float if len(data["values"]) == 0 else None,
    )

    ser.index.name = "timestamp"

    # Output
    return ser


def get_batch_values(
    data_id: str,
    start_date: Union[str, datetime],
    end_date: Union[str, datetime],
    aggregation: str,
    number_of_values: Optional[int] = None,
    aggregation_period: Optional[str] = None,
    aggregation_function: Optional[str] = None,
    batch_type_id: Optional[str] = None,
    batch_index_id: Optional[str] = None,
    unit_id: Optional[str] = None,
    name_data_from: str = "reference",
    append_unit_to_description: bool = True,
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """
    Get values from a batch data from the OIAnalytics API

    Parameters
    ----------
    data_id: str
        The data id on which values are to be retrieved
    start_date: datetime or str
        The beginning of the period to be retrieved
    end_date: datetime or str
        The end of the period to be retrieved
    aggregation: {'RAW_VALUES', 'TIME'}
        How to aggregate the values. If 'TIME', aggregation period and function should be specified.
    number_of_values: int, optional
        If > 0 returns only the number of values specified. Only works for stored data.
    aggregation_period: str, optional
        Required in case 'aggregation' is 'TIME' and is the sampling period of the expected result.
        This period should be an ISO period as described in https://en.wikipedia.org/wiki/ISO_8601#Durations
    aggregation_function: str, optional
        Required in case 'aggregation' is 'TIME' and is the aggregation function to use to aggregate the values within
        the sampling period.
        Possible values are 'FIRST', 'LAST', 'LAST_MINUS_FIRST', 'SUM', 'MIN', 'MAX', 'MEAN', 'MEDIAN', 'STDEV',
        'PERCENTILE5', 'PERCENTILE95', 'DECILE1', 'DECILE9', 'QUARTILE1', 'QUARTILE9', 'COUNT', 'MEAN_MINUS_SIGMA',
        'MEAN_PLUS_SIGMA', 'MEAN_MINUS_TWO_SIGMA', 'MEAN_PLUS_TWO_SIGMA', 'MEAN_MINUS_THREE_SIGMA',
        'MEAN_PLUS_THREE_SIGMA', 'VALUE_CHANGE'.
    batch_type_id: str, optional
        The id of the batch type that should be used to query the values.
        If not present the batch type of the data will be used.
        This batch type should be compatible with the data batch type e.g. it should exist a genealogy between this
        batch type and the batch type of the data.
    batch_index_id: str, optional
        The id of the batch timestamp index that should be used to query the values.
        If not present the batch timestamp will be used. The batch index should belong to the batch type
    unit_id: str, optional
        The id of the unit to use to express the values. If not present a default unit will be used.
        This unit should be compatible with the physical quantity of the data queried.
    name_data_from: {'id', 'reference', 'description'}
        What field should be extracted for naming data.
    append_unit_to_description: bool, default True
        Only used when 'name_data_from' is 'description'. If True, the unit is added after the description.
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    A Series containing the values of the data, indexed by 'batch_id', 'batch_name', 'batch_timestamp'
    """

    # Args validation
    if name_data_from not in ["id", "reference", "description"]:
        raise ValueError(f"Unexpected value for 'name_data_from': {name_data_from}")

    # Query endpoint
    data = endpoints.data.get_batch_values(
        data_id=data_id,
        start_date=start_date,
        end_date=end_date,
        aggregation=aggregation,
        number_of_values=number_of_values,
        aggregation_period=aggregation_period,
        aggregation_function=aggregation_function,
        batch_type_id=batch_type_id,
        batch_index_id=batch_index_id,
        unit_id=unit_id,
        api_credentials=api_credentials,
    )

    # Format series
    if name_data_from == "id":
        data_name = data["data"]["id"]
    elif name_data_from == "reference":
        data_name = data["data"]["reference"]
    elif name_data_from == "description":
        data_name = data["data"]["description"]
        if append_unit_to_description is True:
            data_name = f'{data_name} ({data["unit"]["label"]})'
    else:
        raise ValueError(f"Unexpected value for 'name_data_from': {name_data_from}")

    data_type = data["type"]

    if len(data["values"]) == 0:
        timestamp_index = pd.to_datetime(data["timestamps"]).tz_localize("UTC")
    else:
        timestamp_index = pd.to_datetime(data["timestamps"])

    ser = pd.Series(
        name=data_name,
        index=[
            data["batchIds"],
            data["batchNames"],
            timestamp_index,
        ]
        if data_type == "batch-values"
        else timestamp_index,
        data=data["values"],
        dtype=float if len(data["values"]) == 0 else None,
    )

    if data_type == "batch-values":
        ser.index.names = ["batch_id", "batch_name", "batch_timestamp"]
    else:
        ser.index.name = "timestamp"

    # Output
    return ser


def get_multiple_data_values(
    start_date: Union[str, datetime],
    end_date: Union[str, datetime],
    aggregation: str,
    data_id: Optional[Union[str, List[str]]] = None,
    data_reference: Optional[Union[str, List[str]]] = None,
    number_of_values: Optional[int] = None,
    aggregation_period: Optional[Union[str, timedelta, pd.Timedelta]] = None,
    aggregation_function: Optional[Union[str, List[str]]] = None,
    unit_id: Optional[Union[str, List[str]]] = None,
    unit_label: Optional[Union[str, List[str]]] = None,
    name_data_from: str = "reference",
    append_unit_to_description: bool = True,
    join_series_on: Optional[str] = "index",
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """
    Get values from multiple data at once from the OIAnalytics API

    Parameters
    ----------
    start_date: datetime or str
        The beginning of the period to be retrieved
    end_date: datetime or str
        The end of the period to be retrieved
    aggregation: {'RAW_VALUES', 'TIME'}
        How to aggregate the values. If 'TIME', aggregation period and function should be specified.
    data_id: str or list of str, optional
        The array of data id to query. Required if 'data_reference' is None.
    data_reference: str or list of str, optional
        The array of data reference to query. Required if 'data_id' is None.
    number_of_values: int, optional
        If > 0 returns only the number of values specified. Only works for stored data.
    aggregation_period: str, optional
        Required in case 'aggregation' is 'TIME' and is the sampling period of the expected result.
        This period should be an ISO period as described in https://en.wikipedia.org/wiki/ISO_8601#Durations
    aggregation_function: str, optional
        Required in case 'aggregation' is 'TIME' and is the aggregation function to use to aggregate the values within
        the sampling period.
        Possible values are 'FIRST', 'LAST', 'LAST_MINUS_FIRST', 'SUM', 'MIN', 'MAX', 'MEAN', 'MEDIAN', 'STDEV',
        'PERCENTILE5', 'PERCENTILE95', 'DECILE1', 'DECILE9', 'QUARTILE1', 'QUARTILE9', 'COUNT', 'MEAN_MINUS_SIGMA',
        'MEAN_PLUS_SIGMA', 'MEAN_MINUS_TWO_SIGMA', 'MEAN_PLUS_TWO_SIGMA', 'MEAN_MINUS_THREE_SIGMA',
        'MEAN_PLUS_THREE_SIGMA', 'VALUE_CHANGE'.
    unit_id: str or list of str, optional
        The array of unit id to use to express the values. If not present a default unit will be used.
        All units should be compatible with the physical quantity of the data queried.
        If provided should be the same size as the 'data_id' or 'data_reference' array.
        It cannot be used in conjunction with 'unit_label'.
    unit_label: str or list of str, optional
        The array of unit label to use to express the values. If not present a default unit will be used.
        All units should be compatible with the physical quantity of the data queried.
        If provided should be the same size as the data id or data reference array.
        It cannot be used in conjunction with 'unit_id'.
    name_data_from: {'id', 'reference', 'description'}
        What field should be extracted for naming data.
    append_unit_to_description: bool, default True
        Only used when 'name_data_from' is 'description'. If True, the unit is added after the description.
    join_series_on: {'index', 'timestamp', None}, default 'index'
        Joining strategy for data. 'index' will only work if data is homogeneous (i.e. only time values, or batch values
        on a single batch type). If 'timestamp', only the time index will be kept.
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    If 'join_series_on' is None, each data is returned in a separate Series stored in a dictionary indexed by the data
    name.
    If 'join_series_on' is 'index' (default behaviour) or 'timestamp', a single DataFrame is returned, containing all
    the data in columns.
    """

    # Args validation
    if name_data_from not in ["id", "reference", "description"]:
        raise ValueError(f"Unexpected value for 'name_data_from': {name_data_from}")

    if join_series_on not in ["index", "timestamp", None]:
        raise ValueError(f"Unexpected value for 'join_series_on': {join_series_on}")

    # Query endpoint
    data_list = endpoints.data.get_multiple_data_values(
        start_date=start_date,
        end_date=end_date,
        aggregation=aggregation,
        data_id=data_id,
        data_reference=data_reference,
        number_of_values=number_of_values,
        aggregation_period=aggregation_period,
        aggregation_function=aggregation_function,
        unit_id=unit_id,
        unit_label=unit_label,
        api_credentials=api_credentials,
    )

    # Format series
    batch_types = {}
    data_series = {}
    for data in data_list:
        # Format series
        if name_data_from == "id":
            data_name = data["data"]["id"]
        elif name_data_from == "reference":
            data_name = data["data"]["reference"]
        elif name_data_from == "description":
            data_name = data["data"]["description"]
            if append_unit_to_description is True:
                data_name = f'{data_name} ({data["unit"]["label"]})'
        else:
            raise ValueError(f"Unexpected value for 'name_data_from': {name_data_from}")

        data_type = data["type"]
        batch_types[data_name] = (
            None if data_type == "time-values" else data["batchType"]["id"]
        )

        if len(data["values"]) == 0:
            timestamp_index = pd.to_datetime(data["timestamps"]).tz_localize("UTC")
        else:
            timestamp_index = pd.to_datetime(data["timestamps"])

        ser = pd.Series(
            name=data_name,
            index=[
                data["batchIds"],
                data["batchNames"],
                timestamp_index,
            ]
            if data_type == "batch-values" and join_series_on != "timestamp"
            else timestamp_index,
            data=data["values"],
            dtype=float if len(data["values"]) == 0 else None,
        )

        if data_type == "batch-values" and join_series_on != "timestamp":
            ser.index.names = ["batch_id", "batch_name", "batch_timestamp"]
        else:
            ser.index.name = "timestamp"

        # Store results
        data_series[data_name] = ser

    # Join series
    if join_series_on == "index":
        n_batch_types = pd.Series(batch_types).nunique(dropna=False)
        if n_batch_types > 1:
            raise ValueError(
                "'join_series_on' cannot be set to 'index' on heterogeneous data"
            )
        data_series = pd.DataFrame(data_series)

    elif join_series_on == "timestamp":
        data_series = pd.DataFrame(data_series)

    # Output
    return data_series


def insert_time_values(
    data: Union[pd.Series, pd.DataFrame],
    units: Optional[dict] = None,
    use_external_reference: bool = False,
    timestamp_index_name: str = "timestamp",
    api_credentials: Optional[_credentials.OIAnalyticsAPICredentials] = None,
):
    """
    Insert time values stored in a Series or DataFrame through the OIAnalytics API

    Parameters
    ----------
    data: Series or DataFrame
        The time values to be sent to OIAnalytics. The index should be named 'timestamp'.
    units: dict, optional
        A dictionary indexed by data reference, specifying the values in which it is sent.
    use_external_reference: bool, default False
        Whether or not the data are named using their external reference (else the OIAnalytics reference is used).
    timestamp_index_name: str, default 'timestamp'
        Name of the index level containing the timestamp
    api_credentials: OIAnalyticsAPICredentials, optional
        The credentials to use to query the API. If None, previously set default credentials are used.

    Returns
    -------
    A dictionary of the response from the API, containing the data insert report
    """

    # Init
    if units is None:
        units = {}

    if isinstance(data, pd.Series):
        data = pd.DataFrame(data)
    else:
        data = data.copy()

    # Build DTO
    payload = []

    try:
        data.index = (
            data.index.get_level_values(timestamp_index_name)
            .to_series()
            .apply(utils.get_zulu_isoformat)
        )
    except Exception:
        raise ValueError(
            "The series or dataframe must have an index level named 'timestamp' containing datetime-like values"
        )

    for col in data.columns:
        ser = data[col]
        ser.name = "value"

        # Build payload for the data
        col_dict = {"dataReference": col}
        col_dict["unit"] = units.get(col, None)
        col_dict["values"] = (
            pd.DataFrame(ser).reset_index().dropna().to_dict(orient="records")
        )

        payload.append(col_dict)

    # Query endpoint
    response = endpoints.data.insert_time_values(
        data=payload,
        use_external_reference=use_external_reference,
        api_credentials=api_credentials,
    )

    # Output
    return response



from .single_observation import get_single_obs_time_data, get_single_obs_batch_data
from .files import get_resource_file

__all__ = [
    "get_single_obs_time_data",
    "get_single_obs_batch_data",
    "get_resource_file",
]
